# -*- encoding:utf-8 -*-
"""
    技术线对象，对外执行，输出模块
"""
import math, itertools
import numpy as np
import pandas as pd
from collections import Iterable
from ultron.kdutils.lazy import LazyFunc
from ultron.ump.core.base import FreezeAttrMixin
from ultron.kdutils.decorator import arr_to_numpy
from ultron.kdutils import regression
from ultron.kdutils import stats as stats_util
from ultron.utilities.logger import kd_logger
from ultron.ump.technical.exectue import *

pd.options.display.precision = 4
# numpy精度浮点数显示4位，不使用科学计数法
np.set_printoptions(precision=4, suppress=True)
"""step_x_to_step函数中序列步长的常数单元值"""
g_step_unit = 10
"""预备颜色序列集，超出序列数量应使用itertools.cycle循环绘制"""
K_PLT_MAP_STYLE = ['r', 'g', 'c', 'k', 'm', 'r', 'y']


class Line(FreezeAttrMixin):
    """技术线封装执行对外操作的对象类"""

    def __init__(self, line, line_name, **kwargs):
        """
        :param line: 技术线可迭代序列，内部会通过arr_to_numpy统一转换numpy
        :param line_name: 技术线名称，str对象
        :param kwargs mean: 外部可选择通过kwargs设置mean，如不设置line.mean()
        :param kwargs std: 外部可选择通过kwargs设置std，如不设置line.std()
        :param kwargs high: 外部可选择通过kwargs设置high，如不设置self.mean + self.std
        :param kwargs low: 外部可选择通过kwargs设置low，如不设置self.mean - self.std
        :param kwargs close: 外部可选择通过kwargs设置close，如不设置line[-1]
        """
        # 把序列的nan进行填充，实际上应该是外面根据数据逻辑把nan进行填充好了再传递进来，这里只能都使用bfill填了
        line = pd.Series(line).fillna(method='bfill')
        self.tl = arr_to_numpy(line)
        self.mean = kwargs.pop('mean', self.tl.mean())
        self.std = kwargs.pop('std', self.tl.std())
        self.high = kwargs.pop('high', self.mean + self.std)
        self.low = kwargs.pop('low', self.mean - self.std)
        self.close = kwargs.pop('close', self.tl[-1])

        self.x = np.arange(0, self.tl.shape[0])
        self.line_name = line_name

        for k, v in kwargs:
            # 需要设置什么都通过kwargs设置进来，不然_freeze后无法设置
            setattr(self, k, v)
        # 需要进行定稿，初始化好就不能动
        self._freeze()

    @LazyFunc
    def score(self):
        """
        被LazyFunc装饰：
        score代表当前技术线值在当前的位置， (self.close - self.low) / (self.high - self.low)
        eg：
            self.high ＝ 100， self.low＝0，self.close＝80
            －> (self.close - self.low) / (self.high - self.low) = 0.8
            即代表当前位置在整体的0.8位置上

        :return: 技术线当前score, 返回值在0-1之间
        """
        if self.high == self.low:
            score = 0.8 if self.close > self.low else 0.2
        else:
            score = (self.close - self.low) / (self.high - self.low)
        return score

    @LazyFunc
    def y_zoom(self):
        """
        被LazyFunc装饰：
        获取对象技术线tl被self.x缩放后的序列y_zoom
        :return: 放后的序列y_zoom
        """
        zoom_factor = self.x.max() / self.tl.max()
        y_zoom = zoom_factor * self.tl
        return y_zoom

    def step_x_to_step(self, step_x):
        """
        针对技术线的时间范围步长选择函数，在create_shift_distance，create_regress_trend_channel，
        show_skeleton_channel等涉及时间步长的函数中用来控制步长范围
        :param step_x: 时间步长控制参数，float
        :return: 最终输出被控制在2-len(self.tl), int
        """
        if step_x <= 0:
            # 不正常step_x规范到正常范围中
            #log_func(
            #    'input step_x={} is error, change to step_x=1'.format(step_x))
            step_x = 1
        # 如果需要调整更细的粒度，调整g_step_unit的值
        step = int(math.floor(len(self.tl) / g_step_unit / step_x))
        # 输出被控制在2-len(self.tl)
        step = len(self.tl) if step > len(self.tl) else step
        step = 2 if step < 2 else step
        return step

    def is_up_trend(self, up_deg_threshold=5):
        """
        判断走势是否符合上升走势：
        1. 判断走势是否可以使用一次拟合进行描述
        2. 如果可以使用1次拟合进行描述，计算一次拟合趋势角度
        3. 如果1次拟合趋势角度 >= up_deg_threshold判定上升
        :param up_deg_threshold: 判定一次拟合趋势角度为上升趋势的阀值角度，默认5
        :return: 是否上升趋势
        """
        valid = regression.valid_poly(self.tl, poly=1)
        if valid:
            deg = regression.calc_regress_deg(self.tl)
            if deg >= up_deg_threshold:
                return True
        return False

    def is_down_trend(self, down_deg_threshold=-5):
        """
        判断走势是否符合下降走势：
        1. 判断走势是否可以使用一次拟合进行描述
        2. 如果可以使用1次拟合进行描述，计算一次拟合趋势角度
        3. 如果1次拟合趋势角度 <= down_deg_threshold判定下降
        :param down_deg_threshold: 判定一次拟合趋势角度为下降趋势的阀值角度，默认－5
        :return: 是否下降趋势
        """
        valid = regression.valid_poly(self.tl, poly=1)
        if valid:
            deg = regression.calc_regress_deg(self.tl)
            if deg <= down_deg_threshold:
                return True
        return False

    def create_support_resistance_pos(self, best_poly=0):
        """
        可视化分析技术线阻力位和支撑位，通过sco.fmin_bfgs寻找阻力位支撑位，阻力位点也是通过sco.fmin_bfgs寻找，
        但是要求传递进来的序列已经是标准化后取反的序列
        eg：
            demean_y = ABuStatsUtil.demean(self.tl)： 首先通过demean将序列去均值
            resistance_y = demean_y * -1 ：阻力位序列要取反
            support_y = demean_y ：支持位序列不需要取反
        :param best_poly: 函数使用者可设置best_poly, 设置后就不使用ABuRegUtil.search_best_poly寻找了,
                          详细阅ABuTLExecute.support_resistance_pos
        :return: (技术线支撑位: support_pos, 技术线阻力位: resistance_pos)
        """
        # 首先通过demean将序列去均值
        demean_y = stats_util.demean(self.tl)
        # 阻力位序列要取反
        resistance_y = demean_y * -1
        # 支持位序列不需要取反
        support_y = demean_y

        # 分析技术线支撑位
        support_pos = support_resistance_pos(self.x,
                                             support_y,
                                             best_poly=best_poly,
                                             label='support pos')
        # 分析技术线阻力位
        resistance_pos = support_resistance_pos(self.x,
                                                resistance_y,
                                                best_poly=best_poly,
                                                label='resistance pos')
        return support_pos, resistance_pos

    def create_support_resistance_select_k(self, best_poly=0):
        """
        可视化分析技术线阻力位和支撑位序列从1-序列个数开始聚类，多个聚类器的方差值进行比较，
        通过方差阀值等方法找到最佳聚类个数，最终得到kmean最佳分类器对象
        :param best_poly: 传递show_support_resistance_pos，
                          函数使用者可设置best_poly, 设置后就不使用ABuRegUtil.search_best_poly寻找了
        :return: upport_est, resistance_est, support_pos, resistance_pos
        """
        # 可视化分析技术线阻力位或者支撑位
        support_pos, resistance_pos = self.create_support_resistance_pos(
            best_poly)

        support_pos = np.array(
            [support_pos, [self.tl[support] for support in support_pos]]).T

        resistance_pos = np.array([
            resistance_pos,
            [self.tl[resistance] for resistance in resistance_pos]
        ]).T

        support_est = None
        if len(support_pos) > 1:
            # 多个聚类器的方差值进行比较，通过方差阀值等方法找到最佳聚类个数，最终得到kmean最佳分类器对象
            # 注意这里的show直接False了
            support_est = select_k_support_resistance(support_pos)

        resistance_est = None
        if len(resistance_pos) > 1:
            # 注意这里的show直接False了
            resistance_est = select_k_support_resistance(resistance_pos)

        return support_est, resistance_est, support_pos, resistance_pos

    def create_support_trend(self, best_poly=0, only_last=False):
        """
        最终对技术线只对阻力位进行绘制

        套接：show_support_resistance_select_k－>support_resistance_predict
             ->ABuTLExecute.plot_support_resistance_trend
        :param best_poly: 传递show_support_resistance_pos，
                          函数使用者可设置best_poly, 设置后就不使用ABuRegUtil.search_best_poly寻找了
        :param only_last: 透传ABuTLExecute.plot_support_resistance_trend，控制只绘制时间序列中最后一个发现的阻力或支撑
        """
        support_est, _, support_pos, _ = self.create_support_resistance_select_k(
            best_poly)
        y_trend_dict = {}
        if support_est is not None:
            support_trend = support_resistance_predict(self.x,
                                                       self.tl,
                                                       support_est,
                                                       support_pos,
                                                       is_support=True)
            y_support_trend = create_support_resistance_trend(
                self.x,
                self.tl,
                support_trend,
                'support trend line',
                only_last=only_last)

            if y_support_trend is not None:
                y_trend_dict['support'] = y_support_trend
        return y_trend_dict

    def create_resistance_trend(self, best_poly=0, only_last=False):
        """
        最终对技术线只对支撑位进行绘制

        套接：show_support_resistance_select_k－>support_resistance_predict
             ->ABuTLExecute.plot_support_resistance_trend
        :param best_poly: 传递show_support_resistance_pos，
                          函数使用者可设置best_poly, 设置后就不使用ABuRegUtil.search_best_poly寻找了
        :param only_last: 透传ABuTLExecute.plot_support_resistance_trend，控制只绘制时间序列中最后一个发现的阻力或支撑
        :param plot_org: 透传ABuTLExecute.plot_support_resistance_trend，控制是否绘制线段还是直线，控制是否绘制线段还是直线，
                         plot_org＝True时绘制线段，否则通过LinearRegression进行
        """
        _, resistance_est, _, resistance_pos = self.create_support_resistance_select_k(
            best_poly)

        y_trend_dict = {}

        if resistance_est is not None:
            resistance_trend = support_resistance_predict(self.x,
                                                          self.tl,
                                                          resistance_est,
                                                          resistance_pos,
                                                          is_support=False)
            y_resistance_trend = create_support_resistance_trend(
                self.x,
                self.tl,
                resistance_trend,
                'resistance trend line',
                only_last=only_last)

            if y_resistance_trend is not None:
                y_trend_dict['resistance'] = y_resistance_trend
        return y_trend_dict

    def create_support_resistance_trend(self, best_poly=0, only_last=False):
        """
        套接：show_support_resistance_select_k－>support_resistance_predict
             ->ABuTLExecute.plot_support_resistance_trend
        最终对技术线阻力位和支撑位进行绘制，注意show参数控制的是中间流程中的可视化，不包括
        最终阻力位和支撑的可视化
        :param best_poly: 传递show_support_resistance_pos，
                          函数使用者可设置best_poly, 设置后就不使用ABuRegUtil.search_best_poly寻找了
        :param only_last: 透传ABuTLExecute.plot_support_resistance_trend，控制只绘制时间序列中最后一个发现的阻力或支撑
        """
        support_est, resistance_est, support_pos, resistance_pos = self.create_support_resistance_select_k(
            best_poly)

        y_trend_dict = {}
        if support_est is not None:
            # FIXME 针对极端没有找到足够绘制支撑阻力位的情况做处理
            support_trend = support_resistance_predict(self.x,
                                                       self.tl,
                                                       support_est,
                                                       support_pos,
                                                       is_support=True)

            y_support_trend = create_support_resistance_trend(
                self.x,
                self.tl,
                support_trend,
                'support trend line',
                only_last=only_last)
            if y_support_trend is not None:
                y_trend_dict['support'] = y_support_trend
        else:
            kd_logger.info('can\'t plot support !')

        if resistance_est is not None:
            resistance_trend = support_resistance_predict(self.x,
                                                          self.tl,
                                                          resistance_est,
                                                          resistance_pos,
                                                          is_support=False)
            y_resistance_trend = create_support_resistance_trend(
                self.x,
                self.tl,
                resistance_trend,
                'resistance trend line',
                only_last=only_last)
            if y_resistance_trend is not None:
                y_trend_dict['resistance'] = y_resistance_trend
        else:
            kd_logger.info('can\'t plot resistance !')
        return y_trend_dict

    def create_shift_distance(self,
                              how=EShiftDistanceHow.shift_distance_close,
                              step_x=1.0):
        """
        可视化技术线'路程位移比'，注意默认使用shift_distance_close对应标准路程点位值定义方法，其它方法对应的
        路程终点点位值使用的计算方法并非得到最准确的'路程位移比'，实现详ABuTLExecute.shift_distance
        :param how: EShiftDistanceHow对象或者callable即外部可自行设置方法，即计算算路程终点点位值使用的计算方法可自定义
        :param step_x: 时间步长控制参数，默认1.0，float
        :return 对每一个金融序列切片进行shift_distance的返回结果序列，即每一个序列中的元素为：
                    h_distance(三角底边距离), v_distance(三角垂直距离),
                    distance(斜边，路程), shift(位移), sd（位移路程比：shift / distance）
                所组成的tuple对象
        """
        # 这里使用了缩放后的y，因为确保路程位移比具有更好的适应性
        y = self.y_zoom
        step = self.step_x_to_step(step_x)

        shift_distance_list = []
        for slice_end, _ in zip(np.arange(step, len(y), step),
                                itertools.cycle(K_PLT_MAP_STYLE)):
            slice_start = slice_end - step
            shift_distance_list.append(
                shift_distance(y,
                               how,
                               slice_start=slice_start,
                               slice_end=slice_end))
        return shift_distance_list

    def create_least_valid_poly(self, zoom=False):
        """
        可视化技术线，检测至少poly次拟合曲线可以代表原始曲线y的走势，
        具体详情ABuRegUtil.least_valid_poly
        :param zoom: 透传least_valid_poly是否缩放x,y
        """
        least, rolling_window, y_fit, y_roll_mean, distance_fit, distance_mean = regression.least_valid_poly(
            self.tl, zoom=zoom)
        return least, rolling_window, y_fit, y_roll_mean, distance_fit, distance_mean

    def create_best_poly(self, zoom=False):
        """
        可视化技术线最优拟合次数，寻找poly（1－100）次多项式拟合回归的趋势曲线可以比较完美的代表原始曲线y的走势，
        具体详情ABuRegUtil.search_best_poly
        :param zoom: 透传search_best_poly是否缩放x,y
        """
        best, rolling_window, y_fit, y_roll_mean, distance_fit, distance_mean = regression.search_best_poly(
            self.tl, zoom=zoom, only_poly=False)
        return best, rolling_window, y_fit, y_roll_mean, distance_fit, distance_mean

    def create_regress_trend_channel(self, step_x=1.0):
        """
        可视化技术线拟合曲线及上下拟合通道曲线，返回三条拟合曲线，组成拟合通道
        :param step_x: 时间步长控制参数，默认1.0，float
        """
        channel_list = []
        y = self.tl
        step = self.step_x_to_step(step_x)
        for slice_end, _ in zip(np.arange(step, len(y), step),
                                itertools.cycle(K_PLT_MAP_STYLE)):
            slice_start = slice_end - step
            slice_arr = y[slice_start:slice_end]
            # 通过regress_trend_channel获取切片段的上中下三段拟合曲线值
            y_below, y_fit, y_above = regress_trend_channel(slice_arr)
            x = self.x[slice_start:slice_end]
            channel_list.append({
                'x': x,
                'y_below': y_below,
                'y_fit': y_fit,
                'y_above': y_above
            })
        return channel_list

    def golden_percents(self, percents=(0.1, 0.9)):
        """
        可视化技术线比例分割的区域
        :param percents: float值或者可迭代序列，默认使用(0.1, 0.9)
        :return:
        """
        if not isinstance(percents, Iterable):
            # 如果不是可迭代序列，添加到list中，便于统一处理
            percents = [percents]
        pts_dict = find_percent_point(percents, self.tl)
        return pts_dict

    def create_golden(self, both_golden=True):
        """
        技术线黄金分割
        :param both_golden: 代表同时可视化两种分割线的计算在一个画布上
        :return:
        """
        if both_golden:
            return self.golden_percents(percents=(0.382, 0.618))
        else:
            return {
                'ex': find_golden_point_ex(self.x, self.tl),
                'basic': find_golden_point(self.x, self.tl)
            }

    def create_skeleton(self, how=ESkeletonHow.skeleton_min, step_x=1.0):
        """
        可视化技术线骨架结构
        :param how: 计算数据序列骨架点位的方法，ESkeletonHow对象或者callable即外部可自行设置方法，
                    即计算数据序列骨架点位的方法可自定义
        :param step_x: 时间步长控制参数，默认1.0，float
        """
        step = self.step_x_to_step(step_x)
        # 每个单位都先画一个点，由两个点连成一条直线形成股价骨架图
        last_pos = None
        # 根据how映射计算数据序列骨架点位的方法
        how_func = skeleton_how(how)

        skeleton_res = []
        for slice_end, color in zip(np.arange(step, len(self.tl), step),
                                    itertools.cycle(K_PLT_MAP_STYLE)):
            slice_start = slice_end - step
            slice_arr = self.tl[slice_start:slice_end]
            if how == ESkeletonHow.skeleton_triangle:
                """
                    三角模式骨架点位：确定取最大值，最小值，第三个点位how_func提供
                    如果np.argmax(arr) > np.argmin(arr)即最大值位置在最小值前面，第三点取序列起点，否则取序列终点
                """
                max_pos = (np.argmax(slice_arr) + slice_start,
                           np.max(slice_arr))
                min_pos = (np.argmin(slice_arr) + slice_start,
                           np.min(slice_arr))
                draw_pos = how_func(slice_arr, slice_start)
                res = {
                    'slice_end': slice_end,
                    'max_pos': max_pos,
                    'min_pos': min_pos,
                    'draw_pos': draw_pos
                }

            else:
                # 其它骨架数据计算方法
                draw_pos = (slice_start, how_func(slice_arr))
                temp_pos = last_pos
                # 将这个步长单位内的最小值点赋予last_pos
                last_pos = draw_pos
                res = {
                    'slice_end': slice_end,
                    'last_pos': temp_pos,
                    'draw_pos': draw_pos
                }

            skeleton_res.append(res)
        return skeleton_res

    def create_skeleton_channel(self, step_x=1.0):
        """
        套接show_skeleton，可视化可视化技术线骨架通道，通道由：
        ESkeletonHow.skeleton_min：下通道，
        ESkeletonHow.skeleton_max：上通道，
        ESkeletonHow.skeleton_mean 中轨通道，组成

        :param with_mean: 是否绘制ESkeletonHow.skeleton_mean 中轨通道，默认True
        :param step_x: 时间步长控制参数，默认1.0，float
        """
        min_res = self.create_skeleton(how=ESkeletonHow.skeleton_min,
                                       step_x=step_x)
        max_res = self.create_skeleton(how=ESkeletonHow.skeleton_max,
                                       step_x=step_x)

        mean_res = self.create_skeleton(how=ESkeletonHow.skeleton_mean,
                                        step_x=step_x)

        return min_res, max_res, mean_res



# -*- encoding:utf-8 -*-
"""
    卖出择时示例因子：n倍atr(止盈止损)择时卖出策略
"""

from ultron.ump.factor.sell.base import FactorSellBase, ESupportDirection


class FactorAtrNStop(FactorSellBase):
    """示例n倍atr(止盈止损)因子"""

    def _init_self(self, **kwargs):
        """
            kwargs中可选参数stop_loss_n: 止损的atr倍数
            kwargs中可选参数stop_win_n: 止盈的atr倍数
        """
        if 'stop_loss_n' in kwargs:
            # 设置止损的atr倍数
            self.stop_loss_n = kwargs['stop_loss_n']
            # 在输出生成的orders_pd中及可视化等等显示的名字
            self.sell_type_extra_loss = '{}:stop_loss={}'.format(
                self.__class__.__name__, self.stop_loss_n)

        if 'stop_win_n' in kwargs:
            # 设置止盈的atr倍数
            self.stop_win_n = kwargs['stop_win_n']
            # 在输出生成的orders_pd中及可视化等等显示的名字
            self.sell_type_extra_win = '{}:stop_win={}'.format(
                self.__class__.__name__, self.stop_win_n)

    def support_direction(self):
        """n倍atr(止盈止损)因子支持两个方向"""
        return [
            ESupportDirection.DIRECTION_CAll.value,
            ESupportDirection.DIRECTION_PUT.value
        ]

    def fit_day(self, today, orders):
        """
        止盈event：截止今天相比买入时的收益 * 买入时的期望方向 > n倍atr
        止损event：截止今天相比买入时的收益 * 买入时的期望方向 < -n倍atr
        :param today: 当前驱动的交易日金融时间序列数据
        :param orders: 买入择时策略中生成的订单序列
        :return:
        """
        for order in orders:
            """
                today.close - order.buy_price：截止今天相比买入时的收益，
                order.expect_direction：买单的方向，收益＊方向＝实际收益
            """
            profit = (today.close - order.buy_price) * order.expect_direction
            # atr常数，示例使用今天的atr21与atr14和作为atr常数，亦可以使用其它组合常量的方式
            stop_base = today.atr21 + today.atr14
            if hasattr(
                    self, 'stop_win_n'
            ) and profit > 0 and profit > self.stop_win_n * stop_base:
                # 满足止盈条件卖出股票, 即收益(profit) > n倍atr
                self.sell_type_extra = self.sell_type_extra_win
                # 由于使用了当天的close价格，所以明天才能卖出
                self.sell_tomorrow(order)

            if hasattr(
                    self, 'stop_loss_n'
            ) and profit < 0 and profit < -self.stop_loss_n * stop_base:
                # 满足止损条件卖出股票, 即收益(profit) < -n倍atr
                self.sell_type_extra = self.sell_type_extra_loss
                order.fit_sell_order(self.today_ind, self)
                # 由于使用了当天的close价格，所以明天才能卖出
                self.sell_tomorrow(order)



# -*- coding: utf-8 -*-
import abc, importlib
import warnings
from distutils.version import LooseVersion
import arrow
import numpy as np
import pandas as pd
try:
    from sklearn import __version__ as sklearn_version
    from sklearn import metrics
except ImportError:
    warnings.warn("not installed sklean,please pip install --upgrade sklearn")
try:
    from xgboost import __version__ as xgbboot_version
except ImportError:
    warnings.warn("not installed xgboost,please pip install --upgrade xgboost")
try:
    from lightgbm import __version__ as lightgbm_version
except ImportError:
    warnings.warn(
        "not installed lightgbm,please pip install --upgrade lightgbm")
try:
    from mlxtend import __version__ as mlxtend_version
except ImportError:
    warnings.warn(
        "not installed lightgbm,please pip install --upgrade mlxtend")

from ultron.factor.utilities import list_eq
from ultron.factor.utilities import encode
from ultron.factor.utilities import decode
from ultron.factor.data.transformer import Transformer


class ModelBase(metaclass=abc.ABCMeta):

    def __init__(self, features=None, fit_target=None):
        if features is not None:
            self.formulas = Transformer(features)
            self.features = self.formulas.names
        else:
            self.features = None

        if fit_target is not None:
            self.fit_target = Transformer(fit_target)
        else:
            self.fit_target = None
        self.impl = None
        self.trained_time = None

    def model_encode(self):
        return encode(self.impl)

    @classmethod
    def model_decode(cls, model_desc):
        return decode(model_desc)

    def __eq__(self, rhs):
        return self.model_encode() == rhs.model_encode() \
               and self.trained_time == rhs.trained_time \
               and list_eq(self.features, rhs.features) \
               and encode(self.formulas) == encode(rhs.formulas) \
               and encode(self.fit_target) == encode(rhs.fit_target)

    def fit(self, x: pd.DataFrame, y: np.ndarray):
        self.impl.fit(x[self.features].values, y.flatten())
        self.trained_time = arrow.now().format("YYYY-MM-DD HH:mm:ss")

    def predict(self, x: pd.DataFrame) -> np.ndarray:
        return self.impl.predict(x[self.features].values)

    def score(self, x: pd.DataFrame, y: np.ndarray) -> float:
        return self.impl.score(x[self.features].values, y)

    def ic(self, x: pd.DataFrame, y: np.ndarray) -> float:
        predict_y = self.impl.predict(x[self.features].values)
        return np.corrcoef(predict_y, y)[0, 1]

    def accuracy(self, x: pd.DataFrame, y: np.ndarray) -> float:
        predict_y = self.impl.predict(x[self.features].values)
        return metrics.accuracy_score(y, predict_y)

    def precision(self,
                  x: pd.DataFrame,
                  y: np.ndarray,
                  average='weighted') -> float:
        predict_y = self.impl.predict(x[self.features].values)
        return metrics.precision_score(y, predict_y, average=average)

    def recall(self,
               x: pd.DataFrame,
               y: np.ndarray,
               average='weighted') -> float:
        predict_y = self.impl.predict(x[self.features].values)
        return metrics.recall_score(y, predict_y, average=average)

    def roc_auc(self,
                x: pd.DataFrame,
                y: np.ndarray,
                average='weighted') -> float:
        predict_y = self.impl.predict(x[self.features].values)
        return metrics.roc_auc_score(y, predict_y, average=average)

    def evs(self, x: pd.DataFrame, y: np.ndarray) -> float:
        predict_y = self.impl.predict(x[self.features].values)
        return metrics.mean_absolute_error(y, predict_y)

    def mae(self, x: pd.DataFrame, y: np.ndarray) -> float:
        predict_y = self.impl.predict(x[self.features].values)
        return metrics.mean_absolute_error(y, predict_y)

    def mse(self, x: pd.DataFrame, y: np.ndarray) -> float:
        predict_y = self.impl.predict(x[self.features].values)
        return metrics.mean_squared_error(y, predict_y)

    def r2_score(self,
                 x: pd.DataFrame,
                 y: np.ndarray,
                 multioutput='uniform_average') -> float:
        predict_y = self.impl.predict(x[self.features].values)
        return metrics.r2_score(y, predict_y, multioutput=multioutput)

    @property
    def device(self):
        return self.impl

    @abc.abstractmethod
    def save(self) -> dict:
        if self.__class__.__module__ == '__main__':
            warnings.warn(
                "model is defined in a main module. The model_name may not be correct."
            )

        model_desc = dict(model_name=self.__class__.__module__ + "." +
                          self.__class__.__name__,
                          language='python',
                          saved_time=arrow.now().format("YYYY-MM-DD HH:mm:ss"),
                          features=list(self.features),
                          trained_time=self.trained_time,
                          desc=self.model_encode(),
                          formulas=encode(self.formulas),
                          fit_target=encode(self.fit_target),
                          internal_model=self.impl.__class__.__module__ + "." +
                          self.impl.__class__.__name__)
        return model_desc

    @classmethod
    @abc.abstractmethod
    def load(cls, model_desc: dict):
        obj_layout = cls()
        obj_layout.features = model_desc['features']
        obj_layout.formulas = decode(model_desc['formulas'])
        obj_layout.trained_time = model_desc['trained_time']
        obj_layout.impl = cls.model_decode(model_desc['desc'])
        if 'fit_target' in model_desc:
            obj_layout.fit_target = decode(model_desc['fit_target'])
        else:
            obj_layout.fit_target = None
        return obj_layout


def create_model_base(party_name=None):
    if not party_name:
        return ModelBase
    else:

        class ExternalLibBase(ModelBase):
            _lib_name = party_name

            def save(self) -> dict:
                model_desc = super().save()
                if self._lib_name == 'sklearn':
                    model_desc[self._lib_name + "_version"] = sklearn_version
                elif self._lib_name == 'xgboost':
                    model_desc[self._lib_name + "_version"] = xgbboot_version
                elif self._lib_name == 'lightgbm':
                    model_desc[self._lib_name + "_version"] = lightgbm_version
                elif self._lib_name == 'mlxtend':
                    model_desc[self._lib_name + "_version"] = mlxtend_version
                else:
                    raise ValueError(
                        "3rd party lib name ({0}) is not recognized".format(
                            self._lib_name))
                return model_desc

            @classmethod
            def load(cls, model_desc: dict):
                obj_layout = super().load(model_desc)

                if cls._lib_name == 'sklearn':
                    current_version = sklearn_version
                elif cls._lib_name == 'xgboost':
                    current_version = xgbboot_version
                elif cls._lib_name == 'lightgbm':
                    current_version = lightgbm_version
                elif cls._lib_name == 'mlxtend':
                    current_version = mlxtend_version
                else:
                    raise ValueError(
                        "3rd party lib name ({0}) is not recognized".format(
                            cls._lib_name))

                if LooseVersion(current_version) < LooseVersion(
                        model_desc[cls._lib_name + "_version"]):
                    warnings.warn(
                        'Current {2} version {0} is lower than the model version {1}. '
                        'Loaded model may work incorrectly.'.format(
                            sklearn_version, model_desc[cls._lib_name],
                            cls._lib_name))
                return obj_layout

        return ExternalLibBase


def load_module(name):
    for ml in ['treemodel', 'linearmodel']:
        module_name = 'ultron.optimize.model.{0}'.format(ml)
        module = importlib.import_module(module_name)
        if name in module.__dict__:
            return importlib.import_module(module_name).__getattribute__(name)
    raise ValueError("{0} not in model".format(name))


# -*- coding: utf-8 -*-
import datetime
import time
import json
from ultron.cluster.central.extern_modules.base_module import BaseModule
from ultron.cluster.central.extern_modules.middleware.module import ModelsSingleton


class Module(BaseModule):

    def __init__(self, name, redis_client):
        super(Module, self).__init__(name, redis_client)
        self._func = {
            'startup_task': self.startup_task,
            'shutoff_task': self.shutoff_task,
            'restart_task': self.restart_task,
            'update_task': self.update_task
        }

    def process_respone(self, result):
        name = result['name']
        opcode = result['opcode']
        self._func[opcode](result)

    def update_task(self, result):
        second_time = time.time()
        task_id = int(second_time * 1000000 +
                      datetime.datetime.now().microsecond)
        wid = result['wid']
        #获取消息通道
        queue = ModelsSingleton().get_queue(wid)
        self._redis_client.hset(queue, str(task_id), json.dumps(result))

    def startup_task(self, result):
        second_time = time.time()
        task_id = int(second_time * 1000000 +
                      datetime.datetime.now().microsecond)
        work_name = result['work_name']
        dir_name = result['dir_name']
        task_info = {
            'name': 'tasks',
            'opcode': 'startup_task',
            'work_name': work_name,
            'dir_name': dir_name
        }
        queue_list = ModelsSingleton().get_all_queue()
        for queue in queue_list:
            self._redis_client.hset(queue, str(task_id), json.dumps(task_info))

    def shutoff_task(self, result):
        second_time = time.time()
        task_id = int(second_time * 1000000 +
                      datetime.datetime.now().microsecond)
        work_name = result['work_name']
        dir_name = result['dir_name']
        task_info = {
            'name': 'tasks',
            'opcode': 'shutoff_task',
            'work_name': work_name,
            'dir_name': dir_name
        }
        queue_list = ModelsSingleton().get_all_queue()
        for queue in queue_list:
            self._redis_client.hset(queue, str(task_id), json.dumps(task_info))

    def restart_task(self, result):
        second_time = time.time()
        task_id = int(second_time * 1000000 +
                      datetime.datetime.now().microsecond)
        work_name = result['work_name']
        dir_name = result['dir_name']
        task_info = {
            'name': 'tasks',
            'opcode': 'restart_task',
            'work_name': work_name,
            'dir_name': dir_name
        }
        queue_list = ModelsSingleton().get_all_queue()
        for queue in queue_list:
            self._redis_client.hset(queue, str(task_id), json.dumps(task_info))



# -*- coding: utf-8 -*-
import numpy as np
from .linear import L1LpOptimizer as _L1LpOptimizer
from .linear import LpOptimizer as _LpOptimizer
from .quadratic import TargetVarianceOptimizer as _TargetVarianceOptimizer

class LPOptimizer:
    def __init__(self,
                 objective: np.array,
                 cons_matrix: np.ndarray,
                 lbound: np.ndarray,
                 ubound: np.ndarray,
                 method: str = "deprecated"):

        self._optimizer = _LpOptimizer(cost=objective,
                                       cons_matrix=cons_matrix,
                                       lower_bound=lbound,
                                       upper_bound=ubound)
        self._x, self._f_eval, self._status = self._optimizer.solver(solver="ECOS")

    def status(self):
        return self._status

    def feval(self):
        return self._f_eval

    def x_value(self):
        return self._x

class L1LPOptimizer:

    def __init__(self,
                 objective: np.array,
                 cons_matrix: np.ndarray,
                 current_pos: np.ndarray,
                 target_turn_over: float,
                 lbound: np.ndarray,
                 ubound: np.ndarray):
        self._optimizer = _L1LpOptimizer(cost=objective,
                                         benchmark=current_pos,
                                         l1norm=target_turn_over,
                                         cons_matrix=cons_matrix,
                                         lower_bound=lbound,
                                         upper_bound=ubound)
        self._x, self._f_eval, self._status = self._optimizer.solver()

    def status(self):
        return self._status

    def feval(self):
        return self._f_eval

    def x_value(self):
        return self._x


class TargetVarianceOptimizer:
    def __init__(self, 
                objective: np.array,
                current_pos: np.array,
                target_turn_over: float,
                target_vol: float,
                cov: np.ndarray,
                lbound: np.ndarray,
                ubound: np.ndarray) -> None:
        self._optimizer = _TargetVarianceOptimizer(cost=objective,
                benchmark=current_pos, l1norm=target_turn_over,
                variance_target=target_vol*target_vol, variance=cov,
                lower_bound=lbound,upper_bound=ubound)
        self._x, self._f_val,self._status = self._optimizer.solver()

    def status(self):
        return self._status
    
    def feval(self):
        return self._f_val
    
    def x_value(self):
        return self._x


# -*- encoding:utf-8 -*-
"""封装ML为业务逻辑层进行规范模块"""
from abc import ABCMeta, abstractmethod
from ultron.ump.core.fixes import six
from ultron.ump.model.estimator import Estimator
import pandas as pd
import numpy as np


class Principles(six.with_metaclass(ABCMeta, object)):
    """封装UltronML的上层具体业务逻辑类"""

    def __init__(self, **kwarg):
        """
        从kwarg中输入数据或者，make_xy中本身生产数据，在做完
        make_xy之后，类变量中一定要有x，y和df，使用UltronML继续
        构造self.fiter
        :param kwarg: 直接透传给make_xy的关键子参数，没有必须的参数
        """
        self.make_xy(**kwarg)
        if not hasattr(self, 'x') or not hasattr(self, 'y') \
                or not hasattr(self, 'df'):
            raise ValueError('make_xy failed! x, y not exist!')
        # noinspection PyUnresolvedReferences
        self.fiter = Estimator(self.x, self.y, self.df)

    @abstractmethod
    def make_xy(self, **kwarg):
        """
        子类需要完成的abstractmethod方法，可以从**kwarg中得到数据
        或者make_xy中本身生产数据，但在make_xy之后，类变量中一定要有
        x，y和df
        """
        pass

    def __getattr__(self, item):
        """
        使用UltronML对象self.fiter做为方法代理:
            return getattr(self.fiter, item)
        即UltronMLPd中可以使用UltronML类对象中任何方法
        """
        if item.startswith('__'):
            # noinspection PyUnresolvedReferences
            return super().__getattr__(item)
        return getattr(self.fiter, item)

    def __call__(self):
        """
        方便外面直接call，不用每次去get
        :return: self.fiter
        """
        return self.fiter


# -*- coding: utf-8 -*-

from lib2to3.pgen2.literals import evalString
from tabnanny import verbose
import arrow
import numpy as np
import pandas as pd
import xgboost as xgb
import lightgbm as lgb

from sklearn.ensemble import RandomForestClassifier as RandomForestClassifierImpl
from sklearn.ensemble import RandomForestRegressor as RandomForestRegressorImpl
from sklearn.ensemble import ExtraTreesClassifier as ExtraTreesClassifierImpl
from sklearn.ensemble import ExtraTreesRegressor as ExtraTreesRegressorImpl
from sklearn.ensemble import BaggingClassifier as BaggingClassifierImpl
from sklearn.ensemble import BaggingRegressor as BaggingRegressorImpl
from sklearn.ensemble import AdaBoostClassifier as AdaBoostClassifierImpl
from sklearn.ensemble import AdaBoostRegressor as AdaBoostRegressorImpl
from sklearn.ensemble import VotingClassifier as VotingClassifierImpl
from sklearn.ensemble import VotingRegressor as VotingRegressorImpl
from sklearn.ensemble import GradientBoostingClassifier as GradientBoostingClassifierImpl
from sklearn.ensemble import GradientBoostingRegressor as GradientBoostingRegressorImpl
from mlxtend.classifier import StackingClassifier as StackingClassifierImpl
from mlxtend.regressor import StackingRegressor as StackingRegressorImpl
from sklearn.tree import DecisionTreeClassifier as DecisionTreeClassifierImpl
from sklearn.tree import DecisionTreeRegressor as DecisionTreeRegressorImpl
from xgboost import XGBClassifier as XGBClassifierImpl
from xgboost import XGBRegressor as XGBRegressorImpl
from lightgbm import LGBMClassifier as LGBMClassifierImpl
from lightgbm import LGBMRegressor as LGBMRegressorImpl
from sklearn.model_selection import train_test_split

from ultron.optimize.model.modelbase import create_model_base


class RandomForestRegressor(create_model_base('sklearn')):

    def __init__(self,
                 n_estimators: int = 100,
                 max_features: str = 'auto',
                 features=None,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = RandomForestRegressorImpl(n_estimators=n_estimators,
                                              max_features=max_features,
                                              **kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class RandomForestClassifier(create_model_base('sklearn')):

    def __init__(self,
                 n_estimators: int = 100,
                 max_features: str = 'auto',
                 features=None,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = RandomForestClassifierImpl(n_estimators=n_estimators,
                                               max_features=max_features,
                                               **kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class ExtraTreesClassifier(create_model_base('sklearn')):

    def __init__(self,
                 n_estimators: int = 100,
                 max_features: str = 'auto',
                 features=None,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = ExtraTreesClassifierImpl(n_estimators=n_estimators,
                                             max_features=max_features,
                                             **kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class ExtraTreesRegressor(create_model_base('sklearn')):

    def __init__(self,
                 n_estimators: int = 100,
                 max_features: str = 'auto',
                 features=None,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = ExtraTreesRegressorImpl(n_estimators=n_estimators,
                                            max_features=max_features,
                                            **kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class BaggingClassifier(create_model_base('sklearn')):

    def __init__(self,
                 n_estimators: int = 100,
                 max_features: float = 1.0,
                 features=None,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = BaggingClassifierImpl(n_estimators=n_estimators,
                                          max_features=max_features,
                                          **kwargs)

    @property
    def importances(self):
        return self.impl.estimators_features_


class BaggingRegressor(create_model_base('sklearn')):

    def __init__(self,
                 n_estimators: int = 100,
                 max_features: float = 1.0,
                 features=None,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = BaggingRegressorImpl(n_estimators=n_estimators,
                                         max_features=max_features,
                                         **kwargs)

    @property
    def importances(self):
        return self.impl.estimators_features_


class AdaBoostClassifier(create_model_base('sklearn')):

    def __init__(self,
                 n_estimators: int = 100,
                 learning_rate: float = 1.0,
                 features=None,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = AdaBoostClassifierImpl(n_estimators=n_estimators,
                                           learning_rate=learning_rate,
                                           **kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class AdaBoostRegressor(create_model_base('sklearn')):

    def __init__(self,
                 n_estimators: int = 100,
                 learning_rate: float = 1.0,
                 features=None,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = AdaBoostRegressorImpl(n_estimators=n_estimators,
                                          learning_rate=learning_rate,
                                          **kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class GradientBoostingClassifier(create_model_base('sklearn')):

    def __init__(self,
                 n_estimators: int = 100,
                 max_features: float = 1.0,
                 learning_rate: float = 0.1,
                 features=None,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = GradientBoostingClassifierImpl(n_estimators=n_estimators,
                                                   max_features=max_features,
                                                   learning_rate=learning_rate,
                                                   **kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class GradientBoostingRegressor(create_model_base('sklearn')):

    def __init__(self,
                 n_estimators: int = 100,
                 max_features: float = 1.0,
                 learning_rate: float = 0.1,
                 features=None,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = GradientBoostingRegressorImpl(n_estimators=n_estimators,
                                                  max_features=max_features,
                                                  learning_rate=learning_rate,
                                                  **kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class DecisionTreeClassifier(create_model_base('sklearn')):

    def __init__(self, features=None, fit_target=None, **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = DecisionTreeClassifierImpl(**kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class DecisionTreeRegressor(create_model_base('sklearn')):

    def __init__(self, features=None, fit_target=None, **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = DecisionTreeRegressorImpl(**kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class VotingRegressor(create_model_base('sklearn')):

    def __init__(self,
                 estimators=[],
                 voting='hard',
                 features=None,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = VotingRegressorImpl(estimators=estimators,
                                        voting=voting,
                                        **kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class VotingClassifier(create_model_base('sklearn')):

    def __init__(self,
                 estimators=[],
                 voting='hard',
                 features=None,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = VotingClassifierImpl(estimators=estimators,
                                         voting=voting,
                                         **kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class StackingClassifier(create_model_base('mlxtend')):

    def __init__(self,
                 classifiers=[],
                 meta_classifier=None,
                 features=None,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = StackingClassifierImpl(classifiers=classifiers,
                                           meta_classifier=meta_classifier,
                                           **kwargs)

    @property
    def importances(self):
        return self.impl.clfs_


class StackingRegressor(create_model_base('mlxtend')):

    def __init__(self,
                 regressors=[],
                 meta_regressor=None,
                 features=None,
                 fit_target=None,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = StackingRegressorImpl(regressors=regressors,
                                          meta_regressor=meta_regressor,
                                          **kwargs)

    @property
    def importances(self):
        return self.impl.clfs_


class LGBMRegressor(create_model_base('lightgbm')):

    def __init__(self,
                 n_estimators: int = 100,
                 learning_rate: float = 0.1,
                 max_depth: int = 3,
                 features=None,
                 fit_target=None,
                 n_jobs: int = 1,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = LGBMRegressorImpl(n_estimators=n_estimators,
                                      max_depth=max_depth,
                                      learning_rate=learning_rate,
                                      n_jobs=n_jobs,
                                      **kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class LGBMClassifier(create_model_base('lightgbm')):

    def __init__(self,
                 n_estimators: int = 100,
                 learning_rate: float = 0.1,
                 max_depth: int = 3,
                 features=None,
                 fit_target=None,
                 n_jobs: int = 1,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = LGBMClassifierImpl(n_estimators=n_estimators,
                                       max_depth=max_depth,
                                       learning_rate=learning_rate,
                                       n_jobs=n_jobs,
                                       **kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class LGBMTrainer(create_model_base('lightgbm')):

    def __init__(self,
                 objective='multiclass',
                 n_estimators: int = 100,
                 learning_rate: float = 0.1,
                 max_depth: int = 3,
                 num_class: int = 2,
                 feature_fraction=0.9,
                 bagging_fraction=0.95,
                 eval_sample=None,
                 early_stopping_rounds=None,
                 features=None,
                 fit_target=None,
                 n_jobs: int = 1,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.params = {
            'objective': objective,
            'learning_rate': learning_rate,
            'max_depth': max_depth,
            'metric': ['multi_error', 'multi_logloss'],
            'feature_fraction': feature_fraction,
            'bagging_fraction': bagging_fraction,
            'n_jobs': n_jobs
        }
        self.eval_sample = eval_sample
        self.num_boost_round = n_estimators
        self.early_stopping_rounds = early_stopping_rounds
        self.impl = None
        self.kwargs = kwargs
        self.trained_time = None

    def fit(self, x: pd.DataFrame, y: np.ndarray):
        if self.eval_sample:
            x_train, x_eval, y_train, y_eval = train_test_split(
                x[self.features].values,
                y,
                test_size=self.eval_sample,
                random_state=42)
            d_train = lgb.Dataset(x_train, y_train)
            d_eval = lgb.Dataset(x_eval, y_eval)
            self.impl = lgb.train(self.params,
                                  d_train,
                                  num_boost_round=self.num_boost_round,
                                  evals=[(d_eval, 'eval')],
                                  verbose_eval=False,
                                  **self.kwargs)
        else:
            d_train = lgb.Dataset(x[self.features].values, y)
            self.impl = lgb.train(self.params,
                                  d_train,
                                  num_boost_round=self.num_boost_round,
                                  **self.kwargs)
            self.trained_time = arrow.now().format("YYYY-MM-DD HH:mm:ss")

    def predict(self, x: pd.DataFrame) -> np.ndarray:
        return self.impl.predict(x[self.features].values)

    @property
    def importances(self):
        imps = self.impl.get_fscore().items()
        imps = sorted(imps, key=lambda x: x[0])
        return list(zip(*imps))[1]


class XGBRegressor(create_model_base('xgboost')):

    def __init__(self,
                 n_estimators: int = 100,
                 learning_rate: float = 0.1,
                 max_depth: int = 3,
                 features=None,
                 fit_target=None,
                 n_jobs: int = 1,
                 missing: float = np.nan,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = XGBRegressorImpl(n_estimators=n_estimators,
                                     learning_rate=learning_rate,
                                     max_depth=max_depth,
                                     n_jobs=n_jobs,
                                     missing=missing,
                                     **kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class XGBClassifier(create_model_base('xgboost')):

    def __init__(self,
                 n_estimators: int = 100,
                 learning_rate: float = 0.1,
                 max_depth: int = 3,
                 features=None,
                 fit_target=None,
                 n_jobs: int = 1,
                 missing: float = np.nan,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = XGBClassifierImpl(n_estimators=n_estimators,
                                      learning_rate=learning_rate,
                                      max_depth=max_depth,
                                      n_jobs=n_jobs,
                                      missing=missing,
                                      **kwargs)

    @property
    def importances(self):
        return self.impl.feature_importances_.tolist()


class XGBTrainer(create_model_base('xgboost')):

    def __init__(self,
                 objective='binary:logistic',
                 booster='gbtree',
                 tree_method='hist',
                 n_estimators: int = 100,
                 learning_rate: float = 0.1,
                 max_depth=3,
                 eval_sample=None,
                 early_stopping_rounds=None,
                 subsample=1.,
                 colsample_bytree=1.,
                 features=None,
                 fit_target=None,
                 random_state: int = 0,
                 n_jobs: int = 1,
                 **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.params = {
            'objective': objective,
            'max_depth': max_depth,
            'eta': learning_rate,
            'booster': booster,
            'tree_method': tree_method,
            'subsample': subsample,
            'colsample_bytree': colsample_bytree,
            'nthread': n_jobs,
            'seed': random_state
        }
        self.eval_sample = eval_sample
        self.num_boost_round = n_estimators
        self.early_stopping_rounds = early_stopping_rounds
        self.impl = None
        self.kwargs = kwargs
        self.trained_time = None

    def fit(self, x: pd.DataFrame, y: np.ndarray):
        if self.eval_sample:
            x_train, x_eval, y_train, y_eval = train_test_split(
                x[self.features].values,
                y,
                test_size=self.eval_sample,
                random_state=42)
            d_train = xgb.DMatrix(x_train, y_train)
            d_eval = xgb.DMatrix(x_eval, y_eval)
            self.impl = xgb.train(params=self.params,
                                  dtrain=d_train,
                                  num_boost_round=self.num_boost_round,
                                  evals=[(d_eval, 'eval')],
                                  verbose_eval=False,
                                  **self.kwargs)
        else:
            d_train = xgb.DMatrix(x[self.features].values, y)
            self.impl = xgb.train(params=self.params,
                                  dtrain=d_train,
                                  num_boost_round=self.num_boost_round,
                                  **self.kwargs)

        self.trained_time = arrow.now().format("YYYY-MM-DD HH:mm:ss")

    def predict(self, x: pd.DataFrame) -> np.ndarray:
        d_predict = xgb.DMatrix(x[self.features].values)
        return self.impl.predict(d_predict)

    @property
    def importances(self):
        imps = self.impl.get_fscore().items()
        imps = sorted(imps, key=lambda x: x[0])
        return list(zip(*imps))[1]


# -*- coding: utf-8 -*-
u"""
Created on 2017-1-3

@author: cheng.li
"""

import datetime as dt
import six
cimport cython
from libc.math cimport floor
from ultron.tradingday.Enums._TimeUnits cimport TimeUnits
from ultron.tradingday.Period cimport Period


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline int _monthLength(int month, bint isLeap):
    return _MonthLeapLength[month - 1] if isLeap else  _MonthLength[month - 1]


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline int _monthOffset(int month, bint isLeap):
    return _MonthLeapOffset[month - 1] if isLeap else _MonthOffset[month - 1]


@cython.boundscheck(False)
@cython.wraparound(False)
cdef Date _advance(Date date, int n, int units):
    cdef int d
    cdef int m
    cdef int y
    cdef int addedYear
    cdef int monthLeft
    cdef int length
    cdef int leapFlag

    if units == TimeUnits.Days or units == TimeUnits.BDays:
        return Date(serialNumber=date.__serialNumber__ + n)
    elif units == TimeUnits.Weeks:
        return Date(serialNumber=date.__serialNumber__ + 7 * n)
    elif units == TimeUnits.Months:
        d = date._day
        m = date._month + n
        y = date._year
        addedYear = int(floor(m / 12.))
        monthLeft = m % 12
        if monthLeft == 0:
            monthLeft = 12
            addedYear -= 1
        y += addedYear
        leapFlag = Date.isLeap(y)

        #require(1900 < y < 2200, ValueError, 'year {0:d} is out of bound. It must be in [1901, 2199]'.format(y))

        length = _MonthLeapLength[monthLeft - 1] if leapFlag else _MonthLength[monthLeft - 1]
        if d > length:
            d = length

        return Date(y, monthLeft, d)
    elif units == TimeUnits.Years:
        d = date._day
        m = date._month
        y = date._year + n
        leapFlag = Date.isLeap(y)

        if y <= 1900 or y >= 2200:
            return Date(1900, 1, 1)

        if d == 29 and m == 2 and not leapFlag:
            d = 28

        return Date(y, m, d)


cdef class Date(object):

    @cython.cdivision(True)
    def __init__(self, int year=0, int month=0, int day=0, int serialNumber=0):
        cdef int leap
        cdef int y
        cdef int m
        cdef int d

        if serialNumber:
            self.__serialNumber__ = serialNumber

            y = (int(serialNumber / 365)) + 1900
            if serialNumber <= _YearOffset[y - 1900]:
                y -= 1

            d = serialNumber - _YearOffset[y - 1900]
            m = int(d / 30) + 1
            leap = Date.isLeap(y)
            while d <= _monthOffset(m, leap):
                m -= 1

            #require(1900 < y < 2200, ValueError, 'year {0:d} is out of bound. It must be in [1901, 2199]'.format(y))
            self._year = y
            self._month = m
            self._day = d - _monthOffset(m, leap)
            return
        elif serialNumber and (year or month or day):
            raise ValueError("When serial number is offered, no year or month or day number should be entered")
        elif not (year and month and day):
            raise ValueError("year: {0}, month: {1}, day: {2} can't be null value included".format(year, month, day))

        self._calculate_date(year, month, day)

    cpdef int dayOfMonth(self):
        return self._day

    cpdef int dayOfYear(self):
        return self.__serialNumber__ - _YearOffset[self.year() - 1900]

    cpdef int year(self):
        return self._year

    cpdef int month(self):
        return self._month

    cpdef int weekday(self):
        cdef int w
        w = self.__serialNumber__ % 7
        return 7 if w == 0 else w

    cpdef toDateTime(self):
        return dt.datetime(self.year(), self.month(), self.dayOfMonth())

    @property
    def serialNumber(self):
        return self.__serialNumber__

    def __richcmp__(Date self, Date other, int op):
        if op == 0:
            return self.__serialNumber__ < other.__serialNumber__
        elif op == 1:
            return self.__serialNumber__ <= other.__serialNumber__
        elif op == 2:
            return self.__serialNumber__ == other.__serialNumber__
        elif op == 3:
            return self.__serialNumber__ != other.__serialNumber__
        elif op == 4:
            return self.__serialNumber__ > other.__serialNumber__
        elif op == 5:
            return self.__serialNumber__ >= other.__serialNumber__

    def __add__(self, period):
        if isinstance(period, Period):
            return _advance(self, period.length(), period.units())
        elif isinstance(period, int):
            return Date(serialNumber=self.__serialNumber__ + period)
        else:
            period = Period(period)
            return _advance(self, period.length(), period.units())

    def __sub__(self, period):
        if isinstance(period, Period):
            return _advance(self, -period.length(), period.units())
        elif isinstance(period, int):
            return Date(serialNumber=self.__serialNumber__ - period)
        elif isinstance(period, Date):
            return self.__serialNumber__ - period.__serialNumber__
        else:
            period = Period(period)
            return _advance(self, -period.length(), period.units())

    def __hash__(self):
        return self.__serialNumber__

    def __str__(self):
        cdef int d
        cdef int m
        cdef int y
        d = self.dayOfMonth()
        m = self.month()
        y = self.year()
        return "{0:d}-{1:02d}-{2:02d}".format(y, m, d)

    def __repr__(self):
        cdef int d
        cdef int m
        cdef int y
        d = self.dayOfMonth()
        m = self.month()
        y = self.year()
        return "Date({0:d}, {1:d}, {2:d})".format(y, m, d)

    @staticmethod
    def minDate():
        return Date(1901, 1, 1)

    @staticmethod
    def maxDate():
        return Date(2199, 12, 31)

    @staticmethod
    def endOfMonth(Date date):
        cdef int m
        cdef int y
        m = date.month()
        y = date.year()
        return Date(y, m, _monthLength(m, Date.isLeap(y)))

    @staticmethod
    def isEndOfMonth(Date date):
        cdef int m
        cdef int y
        m = date.month()
        y = date.year()
        return date.dayOfMonth() == _monthLength(m, Date.isLeap(y))

    @staticmethod
    def nextWeekday(Date date, int dayOfWeek):
        cdef int wd = date.weekday()
        return date + ((7 if wd > dayOfWeek else 0) - wd + dayOfWeek)

    @staticmethod
    def nthWeekday(int nth, int dayOfWeek, int m, int y):
        #require(nth > 0, ValueError, "zeroth day of week in a given (month, year) is undefined")
        #require(nth < 6, ValueError, "no more than 5 weekday in a given (month, year)")

        cdef int skip
        cdef int first

        first = Date(y, m, 1).weekday()
        skip = nth - (1 if dayOfWeek >= first else 0)
        return Date(y, m, (1 + dayOfWeek + skip * 7) - first)

    @staticmethod
    def todaysDate():
        today = dt.date.today()
        return Date(today.year, today.month, today.day)

    @staticmethod
    def isLeap(year):
        return _YearIsLeap[year - 1900]

    @staticmethod
    def fromExcelSerialNumber(serialNumber):
        return Date(serialNumber=serialNumber)

    @staticmethod
    def fromDateTime(dateTime):
        return Date(dateTime.year, dateTime.month, dateTime.day)

    @staticmethod
    def parseISO(dateStr):
        return Date(int(dateStr[0:4]), int(dateStr[5:7]), int(dateStr[8:10]))

    @staticmethod
    def strptime(str dateStr, str dateFormat='%Y-%m-%d'):
        pydt = dt.datetime.strptime(dateStr, dateFormat)
        return Date(pydt.year, pydt.month, pydt.day)

    @staticmethod
    def westernStyle(int day, int month, int year):
        return Date(year, month, day)

    def __deepcopy__(self, memo):
        return Date(self._year, self._month, self._day)

    def __reduce__(self):
        d = {}

        return Date, (self._year, self._month, self._day), d

    def __setstate__(self, state):
        pass

    cdef _calculate_date(self, int year, int month, int day):
        cdef int length
        cdef int offset
        cdef int isLeap

        isLeap = Date.isLeap(year)

        length = _monthLength(month, isLeap)
        offset = _monthOffset(month, isLeap)

        self.__serialNumber__ = day + offset + _YearOffset[year - 1900]
        self._day = day
        self._month = month
        self._year = year

cpdef check_date(date):
    if isinstance(date, six.string_types):
        return Date.parseISO(date)
    else:
        return Date.fromDateTime(date)

cdef bint _YearIsLeap[301]
cdef int _YearOffset[301]

_YearIsLeap[:] = [
    # 1900 is leap in agreement with Excel's bug
    # 1900 is out of valid date range anyway
    # 1900-1909
    True, False, False, False, True, False, False, False, True, False,
    # 1910-1919
    False, False, True, False, False, False, True, False, False, False,
    # 1920-1929
    True, False, False, False, True, False, False, False, True, False,
    # 1930-1939
    False, False, True, False, False, False, True, False, False, False,
    # 1940-1949
    True, False, False, False, True, False, False, False, True, False,
    # 1950-1959
    False, False, True, False, False, False, True, False, False, False,
    # 1960-1969
    True, False, False, False, True, False, False, False, True, False,
    # 1970-1979
    False, False, True, False, False, False, True, False, False, False,
    # 1980-1989
    True, False, False, False, True, False, False, False, True, False,
    # 1990-1999
    False, False, True, False, False, False, True, False, False, False,
    # 2000-2009
    True, False, False, False, True, False, False, False, True, False,
    # 2010-2019
    False, False, True, False, False, False, True, False, False, False,
    # 2020-2029
    True, False, False, False, True, False, False, False, True, False,
    # 2030-2039
    False, False, True, False, False, False, True, False, False, False,
    # 2040-2049
    True, False, False, False, True, False, False, False, True, False,
    # 2050-2059
    False, False, True, False, False, False, True, False, False, False,
    # 2060-2069
    True, False, False, False, True, False, False, False, True, False,
    # 2070-2079
    False, False, True, False, False, False, True, False, False, False,
    # 2080-2089
    True, False, False, False, True, False, False, False, True, False,
    # 2090-2099
    False, False, True, False, False, False, True, False, False, False,
    # 2100-2109
    False, False, False, False, True, False, False, False, True, False,
    # 2110-2119
    False, False, True, False, False, False, True, False, False, False,
    # 2120-2129
    True, False, False, False, True, False, False, False, True, False,
    # 2130-2139
    False, False, True, False, False, False, True, False, False, False,
    # 2140-2149
    True, False, False, False, True, False, False, False, True, False,
    # 2150-2159
    False, False, True, False, False, False, True, False, False, False,
    # 2160-2169
    True, False, False, False, True, False, False, False, True, False,
    # 2170-2179
    False, False, True, False, False, False, True, False, False, False,
    # 2180-2189
    True, False, False, False, True, False, False, False, True, False,
    # 2190-2199
    False, False, True, False, False, False, True, False, False, False,
    # 2200
    False
]

_YearOffset[:] = [
    # 1900-1909
    0, 366, 731, 1096, 1461, 1827, 2192, 2557, 2922, 3288,
    # 1910-1919
    3653, 4018, 4383, 4749, 5114, 5479, 5844, 6210, 6575, 6940,
    # 1920-1929
    7305, 7671, 8036, 8401, 8766, 9132, 9497, 9862, 10227, 10593,
    # 1930-1939
    10958, 11323, 11688, 12054, 12419, 12784, 13149, 13515, 13880, 14245,
    # 1940-1949
    14610, 14976, 15341, 15706, 16071, 16437, 16802, 17167, 17532, 17898,
    # 1950-1959
    18263, 18628, 18993, 19359, 19724, 20089, 20454, 20820, 21185, 21550,
    # 1960-1969
    21915, 22281, 22646, 23011, 23376, 23742, 24107, 24472, 24837, 25203,
    # 1970-1979
    25568, 25933, 26298, 26664, 27029, 27394, 27759, 28125, 28490, 28855,
    # 1980-1989
    29220, 29586, 29951, 30316, 30681, 31047, 31412, 31777, 32142, 32508,
    # 1990-1999
    32873, 33238, 33603, 33969, 34334, 34699, 35064, 35430, 35795, 36160,
    # 2000-2009
    36525, 36891, 37256, 37621, 37986, 38352, 38717, 39082, 39447, 39813,
    # 2010-2019
    40178, 40543, 40908, 41274, 41639, 42004, 42369, 42735, 43100, 43465,
    # 2020-2029
    43830, 44196, 44561, 44926, 45291, 45657, 46022, 46387, 46752, 47118,
    # 2030-2039
    47483, 47848, 48213, 48579, 48944, 49309, 49674, 50040, 50405, 50770,
    # 2040-2049
    51135, 51501, 51866, 52231, 52596, 52962, 53327, 53692, 54057, 54423,
    # 2050-2059
    54788, 55153, 55518, 55884, 56249, 56614, 56979, 57345, 57710, 58075,
    # 2060-2069
    58440, 58806, 59171, 59536, 59901, 60267, 60632, 60997, 61362, 61728,
    # 2070-2079
    62093, 62458, 62823, 63189, 63554, 63919, 64284, 64650, 65015, 65380,
    # 2080-2089
    65745, 66111, 66476, 66841, 67206, 67572, 67937, 68302, 68667, 69033,
    # 2090-2099
    69398, 69763, 70128, 70494, 70859, 71224, 71589, 71955, 72320, 72685,
    # 2100-2109
    73050, 73415, 73780, 74145, 74510, 74876, 75241, 75606, 75971, 76337,
    # 2110-2119
    76702, 77067, 77432, 77798, 78163, 78528, 78893, 79259, 79624, 79989,
    # 2120-2129
    80354, 80720, 81085, 81450, 81815, 82181, 82546, 82911, 83276, 83642,
    # 2130-2139
    84007, 84372, 84737, 85103, 85468, 85833, 86198, 86564, 86929, 87294,
    # 2140-2149
    87659, 88025, 88390, 88755, 89120, 89486, 89851, 90216, 90581, 90947,
    # 2150-2159
    91312, 91677, 92042, 92408, 92773, 93138, 93503, 93869, 94234, 94599,
    # 2160-2169
    94964, 95330, 95695, 96060, 96425, 96791, 97156, 97521, 97886, 98252,
    # 2170-2179
    98617, 98982, 99347, 99713, 100078, 100443, 100808, 101174, 101539, 101904,
    # 2180-2189
    102269, 102635, 103000, 103365, 103730, 104096, 104461, 104826, 105191, 105557,
    # 2190-2199
    105922, 106287, 106652, 107018, 107383, 107748, 108113, 108479, 108844, 109209,
    # 2200
    109574
]

cdef int _MonthLength[12]
cdef int _MonthLeapLength[12]
cdef int _MonthOffset[13]
cdef int _MonthLeapOffset[13]

_MonthLength[:] = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
_MonthLeapLength[:] = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

_MonthOffset[:] = [0, 31, 59, 90, 120, 151,  # Jan - Jun
                   181, 212, 243, 273, 304, 334,  # Jun - Dec
                   365]
_MonthLeapOffset[:] = [0, 31, 60, 91, 121, 152,  # Jan - Jun
                       182, 213, 244, 274, 305, 335,  # Jun - Dec
                       366]



# -*- encoding:utf-8 -*-
import pandas as pd
from sklearn.model_selection import KFold
from ultron.optimize.model.modelbase import load_module


def model_fitness(features, model_name, X, Y, params, default_value, mode,
                  n_splits):
    kf = KFold(n_splits=n_splits, shuffle=False)
    res = []
    for train_index, test_index in kf.split(X):
        x_train = X.iloc[train_index]
        y_train = Y.iloc[train_index].values
        x_test = X.iloc[test_index]
        y_test = Y.iloc[test_index].values
        model = load_module(model_name)(features=features, **params)
        model.fit(x_train, y_train)
        res.append({mode: model.__getattribute__(mode)(x_test, y_test)})
    result = pd.DataFrame(res)
    return default_value if result.empty else result[mode].mean()


# -*- encoding:utf-8 -*-
"""
    买入择时示例因子：动态自适应双均线策略
"""
import math
import numpy as np
from ultron.ump.core.helper import pd_resample
from ultron.ump.factor.buy.base import FactorBuyXD, BuyCallMixin
from ultron.ump.technical.line import Line
from ultron.ump.indicator.ma import calc_ma_from_prices


class DoubleMaBuy(FactorBuyXD, BuyCallMixin):
    """示例买入动态自适应双均线策略"""

    def _init_self(self, **kwargs):
        """
            kwargs中可选参数：fast: 均线快线周期，默认不设置，使用自适应动态快线
            kwargs中可选参数：slow: 均线慢线周期，默认不设置，使用自适应动态慢线
            kwargs中可选参数：resample_max: 动态慢线可设置参数重采样周期最大值，默认100，即动态慢线最大100
            kwargs中可选参数：resample_min: 动态慢线可设置参数重采样周期最小值，默认10，即动态慢线最小10
            kwargs中可选参数：change_threshold：动态慢线可设置参数代表慢线的选取阀值，默认0.12
        """

        # 均线快线周期，默认使用5天均线
        self.ma_fast = kwargs.pop('fast', -1)
        self.dynamic_fast = False
        if self.ma_fast == -1:
            self.ma_fast = 5
            self.dynamic_fast = True

        # 均线慢线周期，默认使用60天均线
        self.ma_slow = kwargs.pop('slow', -1)
        self.dynamic_slow = False
        if self.ma_slow == -1:
            self.ma_slow = 60
            self.dynamic_slow = True
        # 动态慢线可设置参数重采样周期最大值，默认90
        self.resample_max = kwargs.pop('resample_max', 100)
        # 动态慢线可设置参数重采样周期最小值，默认10
        self.resample_min = kwargs.pop('resample_min', 10)
        # 动态慢线可设置参数代表慢线的选取阀值，默认0.12
        self.change_threshold = kwargs.pop('change_threshold', 0.12)

        if self.ma_fast >= self.ma_slow:
            # 慢线周期必须大于快线
            raise ValueError('ma_fast >= self.ma_slow !')

        # xd周期数据需要比ma_slow大一天，这样计算ma就可以拿到今天和昨天两天的ma，用来判断金叉，死叉
        kwargs['xd'] = self.ma_slow + 1
        # 设置好xd后可以直接使用基类针对xd的初始化
        super(DoubleMaBuy, self)._init_self(**kwargs)

        # 在输出生成的orders_pd中显示的名字
        self.factor_name = '{}:fast={},slow={}'.format(self.__class__.__name__,
                                                       self.ma_fast,
                                                       self.ma_slow)

    def _dynamic_calc_fast(self, today):
        """
            根据大盘最近一个月走势震荡程度，动态决策快线的值，规则如下：
            如果大盘最近一个月走势使用：
                    一次拟合可以表达：fast＝slow * 0.05 eg: slow=60->fast=60*0.05=3
                    二次拟合可以表达：fast＝slow * 0.15 eg: slow=60->fast=60*0.15=9
                    三次拟合可以表达：fast＝slow * 0.3 eg: slow=60->fast=60*0.3=18
                    四次及以上拟合可以表达：fast＝slow * 0.5 eg: slow=60->fast=60*0.5=30
        """
        # 策略中拥有self.benchmark，即交易基准对象，AbuBenchmark实例对象，benchmark.kl_pd即对应的市场大盘走势
        benchmark_df = self.benchmark.kl_pd
        # 拿出大盘的今天
        benchmark_today = benchmark_df[benchmark_df.date == today.date]
        if benchmark_today.empty:
            # 默认值为慢线的0.15
            return math.ceil(self.ma_slow * 0.15)

        # 要拿大盘最近一个月的走势，准备切片的start，end
        end_key = int(benchmark_today.iloc[0].key)
        start_key = end_key - 20
        if start_key < 0:
            # 默认值为慢线的0.15
            return math.ceil(self.ma_slow * 0.15)
        # 使用切片切出从今天开始向前20天的数据
        benchmark_month = benchmark_df.set_index('key').loc[start_key:end_key +
                                                            1].reset_index()

        # 通过大盘最近一个月的收盘价格做为参数构造TLine对象
        benchmark_month_line = Line(benchmark_month.close,
                                    'benchmark month line')
        # 计算这个月最少需要几次拟合才能代表走势曲线
        least, _, _, _, _, _ = benchmark_month_line.create_least_valid_poly()
        if least == 1:
            # 一次拟合可以表达：fast＝slow * 0.05 eg: slow=60->fast=60*0.05=3
            return math.ceil(self.ma_slow * 0.05)
        elif least == 2:
            # 二次拟合可以表达：fast＝slow * 0.15 eg: slow=60->fast=60*0.15=9
            return math.ceil(self.ma_slow * 0.15)
        elif least == 3:
            # 三次拟合可以表达：fast＝slow * 0.3 eg: slow=60->fast=60*0.3=18
            return math.ceil(self.ma_slow * 0.3)
        else:
            # 四次及以上拟合可以表达：fast＝slow * 0.5 eg: slow=60->fast=60*0.5=30
            return math.ceil(self.ma_slow * 0.5)

    def _dynamic_calc_slow(self, today):
        """
            动态决策慢线的值，规则如下：

            切片最近一段时间的金融时间序列，对金融时间序列进行变换周期重新采样，
            对重新采样的结果进行pct_change处理，对pct_change序列取abs绝对值，
            对pct_change绝对值序列取平均，即算出重新采样的周期内的平均变化幅度，

            上述的变换周期由10， 15，20，30....进行迭代, 直到计算出第一个重新
            采样的周期内的平均变化幅度 > 0.12的周期做为slow的取值
        """
        last_kl = self.past_today_kl(today=today,
                                     past_day_cnt=self.resample_max)

        if last_kl.empty:
            return 60

        for slow in np.arange(self.resample_min, self.resample_max, 5):
            rule = "{}D".format(slow)
            change = abs(
                pd_resample(last_kl.close, rule,
                            how='mean').pct_change()).mean()
            """
                eg: pd_resample(last_kl.close, rule, how='mean')

                    2014-07-23    249.0728
                    2014-09-03    258.3640
                    2014-10-15    240.8663
                    2014-11-26    220.1552
                    2015-01-07    206.0070
                    2015-02-18    198.0932
                    2015-04-01    217.9791
                    2015-05-13    251.3640
                    2015-06-24    266.4511
                    2015-08-05    244.3334
                    2015-09-16    236.2250
                    2015-10-28    222.0441
                    2015-12-09    222.0574
                    2016-01-20    177.2303
                    2016-03-02    226.8766
                    2016-04-13    230.6000
                    2016-05-25    216.7596
                    2016-07-06    222.6420

                    abs(pd_resample(last_kl.close, rule, how='mean').pct_change())

                    2014-09-03    0.037
                    2014-10-15    0.068
                    2014-11-26    0.086
                    2015-01-07    0.064
                    2015-02-18    0.038
                    2015-04-01    0.100
                    2015-05-13    0.153
                    2015-06-24    0.060
                    2015-08-05    0.083
                    2015-09-16    0.033
                    2015-10-28    0.060
                    2015-12-09    0.000
                    2016-01-20    0.202
                    2016-03-02    0.280
                    2016-04-13    0.016
                    2016-05-25    0.060
                    2016-07-06    0.027

                    abs(pd_resample(last_kl.close, rule, how='mean').pct_change()).mean():

                    0.080
            """
            if change > self.change_threshold:
                """
                    返回第一个大于change_threshold的slow,
                    change_threshold默认为0.12，以周期突破的策略一般需要在0.08以上，0.12是为快线留出套利空间
                """
                return slow
        # 迭代np.arange(min, max, 5)都不符合就返回max
        return self.resample_max

    def fit_month(self, today):
        # fit_month即在回测策略中每一个月执行一次的方法
        if self.dynamic_slow:
            # 一定要先动态算ma_slow，因为动态计算fast依赖slow
            self.ma_slow = self._dynamic_calc_slow(today)
        if self.dynamic_fast:
            # 动态计算快线
            self.ma_fast = self._dynamic_calc_fast(today)
        # 动态重新计算后，改变在输出生成的orders_pd中显示的名字
        self.factor_name = '{}:fast={},slow={}'.format(self.__class__.__name__,
                                                       self.ma_fast,
                                                       self.ma_slow)

    def fit_day(self, today):
        """双均线买入择时因子，信号快线上穿慢行形成金叉做为买入信号"""
        # 计算快线
        fast_line = calc_ma_from_prices(self.xd_kl.close,
                                        int(self.ma_fast),
                                        min_periods=1)
        # 计算慢线
        slow_line = calc_ma_from_prices(self.xd_kl.close,
                                        int(self.ma_slow),
                                        min_periods=1)

        if len(fast_line) >= 2 and len(slow_line) >= 2:
            # 今天的快线值
            fast_today = fast_line[-1]
            # 昨天的快线值
            fast_yesterday = fast_line[-2]
            # 今天的慢线值
            slow_today = slow_line[-1]
            # 昨天的慢线值
            slow_yesterday = slow_line[-2]

            if slow_yesterday >= fast_yesterday and fast_today > slow_today:
                # 快线上穿慢线, 形成买入金叉，使用了今天收盘价格，明天买入
                return self.buy_tomorrow()

    """可以选择是否覆盖FactorBuyXD中的buy_tomorrow来增大交易频率，默认基类中self.skip_days = self.xd降低了频率"""
    # def buy_tomorrow(self):
    #     return self.make_buy_order(self.today_ind)



# -*- encoding:utf-8 -*-
"""
    移动平均线，Moving Average，简称MA，原本的意思是移动平均，由于我们将其制作成线形，所以一般称之为移动平均线，简称均线。
    它是将某一段时间的收盘价之和除以该周期。 比如日线MA5指5天内的收盘价除以5 。
    移动平均线是由著名的美国投资专家Joseph E.Granville（葛兰碧，又译为格兰威尔）于20世纪中期提出来的。
    均线理论是当今应用最普遍的技术指标之一，它帮助交易者确认现有趋势、判断将出现的趋势、发现过度延生即将反转的趋势
"""

from collections import Iterable
import pandas as pd
from enum import Enum
from ultron.ump.indicator.base import g_calc_type, ECalcType
from ultron.ump.core.helper import pd_rolling_mean, pd_ewm_mean
from ultron.ump.core.fixes import six
from ultron.kdutils.decorator import catch_error


class EMACalcType(Enum):
    """计算移动移动平均使用的方法"""
    """简单移动平均线"""
    E_MA_MA = 0
    """加权移动平均线"""
    E_MA_EMA = 1


# noinspection PyUnresolvedReferences
def _calc_ma_from_ta(prices, time_period=10, from_calc=EMACalcType.E_MA_MA):
    """
    使用talib计算ma，即透传talib.MA or talib.EMA计算结果
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param time_period: 移动平均的N值，int
    :param from_calc: EMACalcType enum对象，移动移动平均使用的方法
    """

    import talib
    if isinstance(prices, pd.Series):
        prices = prices.values

    if from_calc == EMACalcType.E_MA_MA:
        ma = talib.MA(prices, timeperiod=time_period)
    else:
        ma = talib.EMA(prices, timeperiod=time_period)
    return ma


def _calc_ma_from_pd(prices, time_period=10, from_calc=EMACalcType.E_MA_MA):
    """
    通过pandas计算ma或者ema
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param time_period: 移动平均的N值，int
    :param from_calc: EMACalcType enum对象，移动移动平均使用的方法
    """

    if isinstance(prices, pd.Series):
        prices = prices.values

    if from_calc == EMACalcType.E_MA_MA:
        ma = pd_rolling_mean(prices,
                             window=time_period,
                             min_periods=time_period)
    else:
        ma = pd_ewm_mean(prices, span=time_period, min_periods=time_period)
    return ma


def calc_ma_from_prices(prices,
                        time_period=10,
                        min_periods=None,
                        from_calc=EMACalcType.E_MA_MA):
    """
    通过pandas计算ma或者ema, 添加min_periods参数
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param time_period: 移动平均的N值，int
    :param min_periods: int，默认None则使用time_period
    :param from_calc: EMACalcType enum对象，移动移动平均使用的方法
    """

    if isinstance(prices, pd.Series):
        prices = prices.values

    min_periods = time_period if min_periods is None else min_periods
    if from_calc == EMACalcType.E_MA_MA:
        ma = pd_rolling_mean(prices,
                             window=time_period,
                             min_periods=min_periods)
    else:
        ma = pd_ewm_mean(prices, span=time_period, min_periods=min_periods)
    return ma


"""通过在ABuNDBase中尝试import talib来统一确定指标计算方式, 外部计算只应该使用calc_ma"""
calc_ma = _calc_ma_from_pd if g_calc_type == ECalcType.E_FROM_PD else _calc_ma_from_ta



# -*- coding: utf-8 -*-
import datetime
import json
import socket
from ultron.cluster.work.extern_modules.base_module import BaseModule

class Module(BaseModule):
    def __init__(self, name, wid, token, redis_client):
        super(Module, self).__init__(name, wid, token ,redis_client)
        self._func = {'login_info':self.login_info,
                     'heart_tick':self.heart_tick}
        self._namespace = 'login'
        self._is_logined = 0

    def login_master(self):
        #若登录成功，则通过WID对应的消息队列发送消息
        login_info = {'name':self._namespace,
                      'opcode':'login_in',
                      'wid':self._wid, 'token':self._token,
                      'ip': socket.gethostbyname(socket.getfqdn(socket.gethostname())),
                      'login_time': datetime.datetime.now().strftime("%Y-%m-%d %H:%S:%M")}
        self._redis_client.hset('ultron:work:login', self._wid + login_info['opcode'], json.dumps(login_info))
    
    def heart_tick(self, respone):
        if self._is_logined == 0 or self._wid is None or self._token is None:
            return
        heart_info = {'name': self._namespace,
                     'opcode': 'heart_tick',
                     'wid':self._wid, 'token':self._token,
                     'update_time': datetime.datetime.now().strftime("%Y-%m-%d %H:%S:%M")}
        self._redis_client.hset('ultron:work:login', self._wid + heart_info['opcode'], json.dumps(heart_info))
        
    def process_respone(self, respone):
        name = respone['name']
        opcode = respone['opcode']
        self._func[opcode](respone)
      
    def login_info(self, respone):
        result = respone['result']
        self._is_logined = 1
        print(result)
        
        


# -*- coding: utf-8 -*-

import functools

from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingAverage
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingDecay
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingMax
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingArgMax
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingMin
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingArgMin
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingRank
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMaximumValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMinimumValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingQuantile
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingAllTrue
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingAnyTrue
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingSum
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingVariance
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingStandardDeviation
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingCountedPositive
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingPositiveAverage
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingRSI
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingLogReturn
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingCorrelation
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingResidue
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingSharp
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingSortino
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingMaxDrawdown
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMovingMovingDrawdown
from ultron.sentry.Analysis.TechnicalAnalysis import SecuritySignValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityAverageValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityXAverageValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityMACDValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecuritySqrtValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityDiffValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecuritySimpleReturnValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityLogReturnValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityExpValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityLogValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityPowValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityAbsValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityAcosValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityAcoshValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityAsinValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityAsinhValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityNormInvValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityCeilValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityFloorValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityRoundValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecuritySigmoidValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis import SecurityTanhValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityShiftedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityDeltaValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityCurrentValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityLatestValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityIIFValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityAddedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecuritySubbedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityMultipliedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityDividedValueHolder

from ultron.sentry.Analysis.CrossSectionValueHolders import CSRankedSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSTopNSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSBottomNSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSTopNPercentileSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSBottomNPercentileSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSAverageSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSAverageAdjustedSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSZScoreSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSFillNASecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSPercentileSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSResidueSecurityValueHolder


def CSRank(x, groups=None):
    return CSRankedSecurityValueHolder(x, groups)


def CSTopN(x, n, groups=None):
    return CSTopNSecurityValueHolder(x, n, groups)


def CSTopNQuantile(x, n, groups=None):
    return CSTopNPercentileSecurityValueHolder(x, n, groups)


def CSBottomN(x, n, groups=None):
    return CSBottomNSecurityValueHolder(x, n, groups)


def CSBottomNQuantile(x, n, groups=None):
    return CSBottomNPercentileSecurityValueHolder(x, n, groups)


def CSMean(x, groups=None):
    return CSAverageSecurityValueHolder(x, groups)


def CSMeanAdjusted(x, groups=None):
    return CSAverageAdjustedSecurityValueHolder(x, groups)


def CSQuantiles(x, groups=None):
    return CSPercentileSecurityValueHolder(x, groups)


def CSZScore(x, groups=None):
    return CSZScoreSecurityValueHolder(x, groups)


def CSFillNA(x, groups=None):
    return CSFillNASecurityValueHolder(x, groups)


def CSRes(left, right):
    return CSResidueSecurityValueHolder(left, right)


def SIGN(x='x'):
    return SecuritySignValueHolder(x)


def AVG(x='x'):
    return SecurityAverageValueHolder(x)


def EMA(window, x='x'):
    return SecurityXAverageValueHolder(window, x)


def MACD(short, long, x='x'):
    return SecurityMACDValueHolder(short, long, x)


def RSI(window, x='x'):
    return SecurityMovingRSI(window, x)


def MARETURNLog(window, x='x'):
    return SecurityMovingLogReturn(window, x)


def MCORR(window, x='x', y='y'):
    return SecurityMovingCorrelation(window, x, y)


def MRes(window, x='x', y='y'):
    return SecurityMovingResidue(window, x, y)


def MSharp(window, x='x', y='y'):
    return SecurityMovingSharp(window, x, y)


def MSortino(window, x='x', y='y'):
    return SecurityMovingSortino(window, x, y)


def MMaxDrawdown(window, x='x'):
    return SecurityMovingMaxDrawdown(window, x)


def MMDrawdown(window, x='x'):
    return SecurityMovingMovingDrawdown(window, x)


def MA(window, x='x'):
    return SecurityMovingAverage(window, x)


def MADecay(window, x='x'):
    return SecurityMovingDecay(window, x)


def MMAX(window, x='x'):
    return SecurityMovingMax(window, x)


def MARGMAX(window, x='x'):
    return SecurityMovingArgMax(window, x)


def MMIN(window, x='x'):
    return SecurityMovingMin(window, x)


def MARGMIN(window, x='x'):
    return SecurityMovingArgMin(window, x)


def MRANK(window, x='x'):
    return SecurityMovingRank(window, x)


def ADDED(x='x', y='y'):
    return SecurityAddedValueHolder(x, y)


def SUBBED(x='x', y='y'):
    return SecuritySubbedValueHolder(x, y)


def MUL(x='x', y='y'):
    return SecurityMultipliedValueHolder(x, y)


def DIV(x='x', y='y'):
    return SecurityDividedValueHolder(x, y)


def MAXIMUM(x='x', y='y'):
    return SecurityMaximumValueHolder(x, y)


def MINIMUM(x='x', y='y'):
    return SecurityMinimumValueHolder(x, y)


def MQUANTILE(window, x='x'):
    return SecurityMovingQuantile(window, x)


def MALLTRUE(window, x='x'):
    return SecurityMovingAllTrue(window, x)


def MANYTRUE(window, x='x'):
    return SecurityMovingAnyTrue(window, x)


def MSUM(window, x='x'):
    return SecurityMovingSum(window, x)


def MVARIANCE(window, x='x'):
    return SecurityMovingVariance(window, x)


def MSTD(window, x='x'):
    return SecurityMovingStandardDeviation(window, x)


def MNPOSITIVE(window, x='x'):
    return SecurityMovingCountedPositive(window, x)


def MAPOSITIVE(window, x='x'):
    return SecurityMovingPositiveAverage(window, x)


def CURRENT(x='x'):
    return SecurityCurrentValueHolder(x)


def LAST(x='x'):
    return SecurityLatestValueHolder(x)


def SQRT(x='x'):
    return SecuritySqrtValueHolder(x)


def DIFF(x='x'):
    return SecurityDiffValueHolder(x)


def RETURNSimple(x='x'):
    return SecuritySimpleReturnValueHolder(x)


def RETURNLog(x='x'):
    return SecurityLogReturnValueHolder(x)


def EXP(x):
    return SecurityExpValueHolder(x)


def LOG(x):
    return SecurityLogValueHolder(x)


def POW(x, n):
    return SecurityPowValueHolder(x, n)


def ABS(x):
    return SecurityAbsValueHolder(x)


def ACOS(x):
    return SecurityAcosValueHolder(x)


def ACOSH(x):
    return SecurityAcoshValueHolder(x)


def ASIN(x):
    return SecurityAsinValueHolder(x)


def ASINH(x):
    return SecurityAsinhValueHolder(x)


def NORMINV(x):
    return SecurityNormInvValueHolder(x)


def NORMINV(x):
    return SecurityNormInvValueHolder(x)


def CEIL(x):
    return SecurityCeilValueHolder(x)


def FLOOR(x):
    return SecurityFloorValueHolder(x)


def ROUND(x):
    return SecurityRoundValueHolder(x)


def SIGMOID(x):
    return SecuritySigmoidValueHolder(x)


def TANH(x):
    return SecurityTanhValueHolder(x)


def SHIFT(n, x):
    return SecurityShiftedValueHolder(n, x)


def DELTA(n, x):
    return SecurityDeltaValueHolder(n, x)


def IIF(flag, left, right):
    return SecurityIIFValueHolder(flag, left, right)


def INDUSTRY(name, level=1):
    name = name.lower()
    if name == 'sw':
        if level in (1, 2, 3):
            return LAST(name + str(level))
        else:
            raise ValueError('{0} is not recognized level for {1}'.format(
                level, name))
    else:
        raise ValueError(
            'currently only `sw` industry is supported. {1} is not a recognized industry type'
        )



# -*- encoding:utf-8 -*-
import functools, os, warnings, time, logging
import numpy as np
import pandas as pd
from collections import Iterable
from ultron.ump.core.fixes import six, signature


def warnings_filter(func):
    """
        作用范围：函数装饰器 (模块函数或者类函数)
        功能：被装饰的函数上的警告不会打印，忽略
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        warnings.simplefilter('ignore')
        ret = func(*args, **kwargs)
        if 'IGNORE_WARNINGS' not in os.environ or not os.environ[
                'IGNORE_WARNINGS']:
            warnings.simplefilter('default')
        return ret

    return wrapper


def first_delegate_has_method(delegate, check_params=True):
    """
    装饰在类函数上，如果delegate有定义对应名称方法，优先使用delegate中的方法，否则使用被装饰的方法

        eg:

        class A:
        def a_func(self):
            print('a.a_func')

        class B:
            def __init__(self):
                self.a = A()

            @DelegateUtil.first_delegate_has_method('a')
            def a_func(self):
                print('b.a_func')

        in: B().a_func()
        out: a.a_func

    :param delegate: str对象，被委托的类属性对象名称，从被装饰的方法的类成员变量中寻找对应名字的对象
    :param check_params: 是否检测方法签名是否相同，默认检测
    """

    def decorate(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if len(args) > 0:
                wrap_self = args[0]
                # 首先从被装饰的类对象实例获取delegate
                delegate_obj = getattr(wrap_self, delegate, None)
                # 被装饰的类对象实例存在且存在和func.__name__一样的方法
                if delegate_obj is not None and hasattr(
                        delegate_obj, func.__name__):
                    # 被委托的对象的函数的方法签名
                    delegate_params = list(
                        signature(getattr(delegate_obj,
                                          func.__name__)).parameters.keys())
                    # 被装饰的函数的方法签名
                    func_params = list(signature(func).parameters.keys())[1:]
                    # print(func_params)
                    # print(delegate_params)
                    # TODO 增加检测规范，如只检测参数order及类型，不全匹配名字
                    if not check_params or delegate_params == func_params:
                        # 一致就优先使用被委托的对象的同名函数
                        return getattr(delegate_obj, func.__name__)(*args[1:],
                                                                    **kwargs)
            return func(*args, **kwargs)

        return wrapper

    return decorate


def replace_word_delegate_has_method(delegate,
                                     key_word,
                                     replace_word,
                                     check_params=True):
    """
    不在delegate中寻找完全一样的方法名字，在被装饰的方法名字中的key_word替换为replace_word后再在delegate中寻找，找到优先使用
    否则继续使用被装饰的方法
        eg:
            class A:
                def a_func(self):
                    print('a.a_func')

            class B:
                def __init__(self):
                    self.a = A()

                @DelegateUtil.replace_word_delegate_has_method('a', key_word='b', replace_word='a')
                def b_func(self):
                    print('b.b_func')

            in: B().b_func()
            out: a.a_func

    :param delegate: str对象，被委托的类属性对象名称，从被装饰的方法的类成员变量中寻找对应名字的对象
    :param key_word: 被装饰的函数名称中将被replace_word替换的key_word，str对象
    :param replace_word: 替换key_word形成要寻找的在被委托函数中的名字，str对象
    :param check_params: 是否检测方法签名是否相同，默认检测
    """

    def decorate(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if len(args) > 0:
                wrap_self = args[0]
                # 首先从被装饰的类对象实例获取delegate
                delegate_obj = getattr(wrap_self, delegate, None)
                if delegate_obj is not None:
                    #  被装饰的类对象实例存在
                    org_func_name = func.__name__
                    if len(replace_word) > 0 and key_word in org_func_name:
                        # 使用replace_word替换原始函数名称org_func_name中的key_word
                        delegate_func_name = org_func_name.replace(
                            key_word, replace_word)
                    else:
                        delegate_func_name = org_func_name

                    if hasattr(delegate_obj, delegate_func_name):
                        # 被装饰的类对象中确实存在delegate_func_name
                        delegate_params = list(
                            signature(
                                getattr(delegate_obj,
                                        delegate_func_name)).parameters.keys())
                        func_params = list(
                            signature(func).parameters.keys())[1:]
                        # TODO 增加检测规范，如只检测参数order及类型，不全匹配名字
                        if not check_params or delegate_params == func_params:
                            # 参数命名一致就优先使用被委托的对象的函数
                            return getattr(delegate_obj,
                                           delegate_func_name)(*args[1:],
                                                               **kwargs)
            return func(*args, **kwargs)

        return wrapper

    return decorate


def params_to_numpy(func):
    """
        函数装饰器：不定参数装饰器，定参数转换使用ABuScalerUtil中的装饰器arr_to_numpy(func)
        将被装饰函数中的参数中所有可以迭代的序列转换为np.array
    """

    @functools.wraps(func)
    def wrapper(*arg, **kwargs):
        # 把arg中的可迭代序列转换为np.array
        arg_list = [arr_to_numpy(param) for param in arg]
        # 把kwargs中的可迭代序列转换为np.array
        arg_dict = {
            param_key: arr_to_numpy(kwargs[param_key])
            for param_key in kwargs
        }
        return func(*arg_list, **arg_dict)

    return wrapper


def arr_to_numpy(arr):
    """
        函数装饰器：将可以迭代的序列转换为np.array，支持pd.DataFrame或者pd.Series
        ，list，dict, list，set，嵌套可迭代序列, 混嵌套可迭代序列
    """
    # TODO Iterable和six.string_types的判断抽出来放在一个模块，做为Iterable的判断来使用
    if not isinstance(arr, Iterable) or isinstance(arr, six.string_types):
        return arr

    if not isinstance(arr, np.ndarray):
        if isinstance(arr, pd.DataFrame) or isinstance(arr, pd.Series):
            # 如果是pandas直接拿values
            arr = arr.values
        elif isinstance(arr, dict):
            # 针对dict转换np.array
            arr = np.array(list(arr.values())).T
        else:
            arr = np.array(arr)
    return arr


def consume_time(func):
    """
    作用范围：函数装饰器 (模块函数或者类函数)
    功能：简单统计被装饰函数运行时间
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        print('{} cost {}s'.format(func.__name__,
                                   round(end_time - start_time, 3)))
        return result

    return wrapper


def singleton(cls):
    """
        作用范围：类装饰器
        功能：被装饰后类变成单例类
    """

    instances = {}

    @functools.wraps(cls)
    def get_instance(*args, **kw):
        if cls not in instances:
            # 不存在实例instances才进行构造
            instances[cls] = cls(*args, **kw)
        return instances[cls]

    return get_instance


def catch_error(return_val=None, log=True):
    """
    作用范围：函数装饰器 (模块函数或者类函数)
    功能：捕获被装饰的函数中所有异常，即忽略函数中所有的问题，用在函数的执行级别低，且不需要后续处理
    :param return_val: 异常后返回的值，
                eg:
                    class A:
                        @ABuDTUtil.catch_error(return_val=100)
                        def a_func(self):
                            raise ValueError('catch_error')
                            return 100
                    in: A().a_func()
                    out: 100
    :param log: 是否打印错误日志
    """

    def decorate(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                logging.exception(e) if log else logging.debug(e)
                return return_val

        return wrapper

    return decorate



# -*- coding: utf-8 -*-

from ultron.tradingday.Enums._Months cimport Months as mt


cpdef enum Months:
    January = mt.January
    February = mt.February
    March = mt.March
    April = mt.April
    May = mt.May
    June = mt.June
    July = mt.July
    August = mt.August
    September = mt.September
    October = mt.October
    November = mt.November
    December = mt.December


# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import gevent.monkey
import multiprocessing
import itertools, time
from .accumulators import mutated_pool, cross_pool, dependency, calc_new_factor

GLOBAL_ORDER_ID = 0


class GeneticCrossFactors(object):

    def __init__(self,
                 del_prob,
                 add_prob,
                 cross_prob,
                 change_prob,
                 conver_prob,
                 cross_rate=0.01,
                 change_rate=0.01,
                 generation=500,
                 group_num=10,
                 parallel=None,
                 objective=None):
        #ori:原始特征，主要用于计算时候获取原始因子  new:变异特征
        self._del_prob = del_prob  # 删除原始特征概率
        self._add_prob = add_prob  # 添加新特征概率
        self._cross_prob = cross_prob  #交叉概率
        self._change_prob = change_prob  #变异概率
        self._conver_prob = conver_prob  # 收敛差值
        self._cross_rate = cross_rate  # 交叉比例
        self._change_rate = change_rate  # 变异比例
        self._generations = generation
        self._group_num = group_num
        self._objective = objective
        self._parallel = parallel if parallel is not None else multiprocessing.cpu_count(
        )

    # 初代种群生成
    def ga_generate_ori(self, ori_field, add_field, evalue_cols):
        ## 选取原始特征过程
        ori_list = (np.random.uniform(0, 1, (1, len(ori_field))) >
                    self._del_prob).tolist()[0]
        add_list = (np.random.uniform(0, 1, (1, len(add_field))) <
                    self._add_prob).tolist()[0]
        new_list = ori_list + add_list
        new_list = [1 if i > 0 else 0 for i in new_list]
        new_array = evalue_cols[np.array(
            [True if i > 0 else False for i in new_list])]
        #return {'ori':new_array, 'new':new_array}
        return new_array

## 交叉变异过程

    def ga_cross_next_group(self,
                            ori_group,
                            ori_columns,
                            dict_score='ori',
                            generation=0):
        new_dict = ori_group.copy()
        if generation == 0:
            score = 1.0 / len(ori_group)
            dict_score = {k: score for k in ori_group.keys()}
        g, p = np.array([[k, v] for k, v in dict_score.items()]).T
        cross_group = np.random.choice(g,
                                       size=int(len(g) / 2),
                                       p=p,
                                       replace=False)
        for (fa, mo) in itertools.combinations(cross_group, 2):
            global GLOBAL_ORDER_ID
            GLOBAL_ORDER_ID += 1
            flag = GLOBAL_ORDER_ID
            fa_code, mo_code = ori_group[fa], ori_group[mo]
            cut_point = np.random.randint(1, len(fa_code) / 2)
            fa_code0, fa_code1 = fa_code[:cut_point], fa_code[cut_point:]
            mo_code0, mo_code1 = mo_code[:cut_point], mo_code[cut_point:]
            new1 = np.hstack([fa_code0, mo_code1])  #前一半父类基因，后一半母类基因
            #父类母类中存在共用基因，只保留一个基因
            new1 = np.array(list(set(new1)))

            #加入原始基因增加遗传多样性
            ori_prob = np.random.uniform(0, 1)
            if ori_prob < 0.2:
                section_columns = set(ori_columns) - set(new1)
                new_code = np.array(list(section_columns))[np.random.randint(
                    0, len(section_columns),
                    int(len(section_columns) * 0.01 + 1))]
                pos = np.random.randint(0, len(new1), len(new_code))
                new1[pos] = new_code  #前一半父类基因，后一半母类基因, 最后一半加入原始基因

            new_group = list(new1.copy())
            #count = int(len(new_group) * self._cross_prob)
            #交叉
            prob = np.random.uniform(0, 1)
            if prob < self._cross_prob:
                count = int(len(new_group) * self._cross_rate + 1)
                for i in range(0, count):
                    fa_pos = np.random.randint(0, len(new1) / 2)  # 取为1
                    mo_pos = np.random.randint(len(new1) / 2, len(new1))  # 取为1
                    accumulators = cross_pool[int(
                        np.random.uniform(1,
                                          len(cross_pool) - 1))]
                    #判断是否为原始
                    if new1[fa_pos] in ori_columns and new1[
                            mo_pos] in ori_columns:
                        new_factor = accumulators.__name__ + '(\'' + new1[
                            fa_pos] + '\',\'' + new1[mo_pos] + '\')'
                    elif new1[fa_pos] in ori_columns and new1[
                            mo_pos] not in ori_columns:
                        new_factor = accumulators.__name__ + '(\'' + new1[
                            fa_pos] + '\',' + new1[mo_pos] + ')'
                    elif new1[fa_pos] not in ori_columns and new1[
                            mo_pos] in ori_columns:
                        new_factor = accumulators.__name__ + '(' + new1[
                            fa_pos] + ',\'' + new1[mo_pos] + '\')'
                    elif new1[fa_pos] not in ori_columns and new1[
                            mo_pos] not in ori_columns:
                        new_factor = accumulators.__name__ + '(' + new1[
                            fa_pos] + ',' + new1[mo_pos] + ')'
                    new_group.append(new_factor)
            #变异
            prob = np.random.uniform(0, 1)
            if prob < self._change_prob:
                count = int(len(new_group) * self._change_rate + 1)
                for i in range(0, count):
                    pos = np.random.randint(0, len(new1))  #交叉生成的不变异
                    accumulators = mutated_pool[int(
                        np.random.uniform(2,
                                          len(mutated_pool) - 1))]
                    #判断是否为原始
                    if new1[pos] in ori_columns:
                        new_factor = accumulators.__name__ + '(\'' + new1[
                            pos] + '\')'
                    else:
                        new_factor = accumulators.__name__ + '(' + new1[
                            pos] + ')'
                    new_group[pos] = new_factor
            new_dict[flag] = np.array(new_group)
        return new_dict

    #种群特征计算
    def calc_evalue_group(self, params):
        sub_group, evalue_cols, total_data, diff_filed, g = params
        #sub_factor_data = total_data.set_index(['trade_date','code'])[sub_group['ori']].T
        #自身变异算法使用apply计算
        mutated_cross_columns = set(sub_group) - set(
            total_data.columns)  #交叉变异列
        if len(mutated_cross_columns) > 0:
            result = dependency(mutated_cross_columns)
            factor_columns = ','.join(result.values()).split(',')
            sub_factor_data = total_data.set_index(
                ['trade_date'])[factor_columns + ['code']]
            news_factor_data = calc_new_factor(mutated_cross_columns,
                                               sub_factor_data)
            surplus = set(sub_group) - mutated_cross_columns
            #复制剩余无需更改的
            for ori in list(surplus):
                news_factor_data[ori] = total_data[ori].values
        else:
            news_factor_data = total_data[list(sub_group)].copy()
        #复制其他项目
        for diff in diff_filed:
            news_factor_data[diff] = total_data[diff].values
        score = self._objective(
            news_factor_data.replace([np.inf, -np.inf], np.nan), sub_group)
        if score == np.nan: score = 0
        return score, sub_group, g

    ## 种群个体能力评价
    def ga_evalue_group(self, sub_group, total_data, evalue_cols, diff_filed):
        jobs = []
        score_dict = {}
        cols_dict = {}
        start_time = time.time()
        '''
        for key in sub_group.keys():
            #切换多进程计算
            score, new_group, g = self.calc_evalue_group([sub_group[key], evalue_cols, total_data, 
                                     diff_filed,key])
            score_dict[g] = score
            cols_dict[g] = new_group
        '''
        cpus_count = multiprocessing.cpu_count(
        ) if multiprocessing.cpu_count() < len(sub_group) else len(sub_group)
        for key in sub_group.keys():
            jobs.append(
                [sub_group[key], evalue_cols, total_data, diff_filed, key])
        with multiprocessing.Pool(processes=cpus_count) as p:
            values_list = p.map(self.calc_evalue_group, jobs)
        for values in values_list:
            score = values[0]
            sub_group = values[1]
            g = values[2]
            score_dict[g] = score
            cols_dict[g] = sub_group
        print(time.time() - start_time)
        return score_dict, cols_dict

    #种群变异
    def ga_kill_group(self,
                      total_data,
                      ori_group,
                      dict_score,
                      evalue_cols,
                      diff_filed,
                      generation=0):
        # 生成新一代特征
        ori_columns = list(set(total_data.columns) - set(diff_filed))
        sub_group = self.ga_cross_next_group(ori_group,
                                             ori_columns,
                                             dict_score=dict_score,
                                             generation=generation)
        ori_score_dict, cols_dict = self.ga_evalue_group(
            sub_group, total_data, evalue_cols, diff_filed)
        ori_score_se = pd.Series(ori_score_dict)
        print('种群数%d, 种群均分:%f' % (len(ori_score_se), ori_score_se.mean()))
        #真实分数
        fact_se = ori_score_se.sort_values(ascending=False)[:self._group_num]
        #比例分数
        score_se = ori_score_se.sort_values(
            ascending=False)[:self._group_num] / (ori_score_se.sort_values(
                ascending=False)[:self._group_num].sum())
        liv_group = {}
        for k in score_se.keys():
            liv_group[k] = cols_dict[k]
        return score_se, dict(fact_se), liv_group

    def genetic_run(self,
                    total_data,
                    diff_filed,
                    sequence=['trade_date', 'code'],
                    factors_filed=None,
                    strong_field=None,
                    weak_field=None,
                    is_best=True):
        if factors_filed is not None:
            point = int(np.random.uniform(0, len(factors_filed)) / 2)
            ori_field = factors_filed[:point]
            add_field = factors_filed[point:]
        elif strong_field is not None and weak_field is not None:
            ori_field = strong_field
            add_field = weak_field

        total_data = total_data.sort_values(by=sequence, ascending=True)
        evalue_cols = np.array(ori_field + add_field)
        last_score = 0.0
        #初始种群
        ori_group = {
            i: self.ga_generate_ori(ori_field, add_field, evalue_cols)
            for i in range(self._group_num)
        }
        global GLOBAL_ORDER_ID
        GLOBAL_ORDER_ID += (len(ori_group) - 1)
        for i in range(self._generations):
            if i == 0:
                score_se, fact_se, liv_group = self.ga_kill_group(
                    total_data, ori_group, 'ori', evalue_cols, diff_filed, 0)
            else:
                score_se, fact_se, liv_group = self.ga_kill_group(
                    total_data, liv_group, score_se, evalue_cols, diff_filed,
                    0)

            best_fact_score = pd.Series(fact_se).sort_values(
                ascending=False)[0:1]
            best_score = pd.Series(score_se).sort_values(ascending=False)
            diff_fact_score = best_fact_score.values[0] - last_score
            diff_score = best_score[0:1].values[0] - best_score[-1:].values[0]

            now_best_score = (best_fact_score.values[0]
                              if best_fact_score.values[0] > last_score else
                              last_score)
            print('繁衍代数:%d,最好分数:%f,%d(%f)和%d(%f)最好组分数差值%f,%d代最好最差种群分数差值%f' %
                  ((i + 1), now_best_score, i + 1, best_fact_score.values[0],
                   i, last_score, diff_fact_score, i + 1, diff_score))
            best_code = list(liv_group.keys())[0]
            best_group = liv_group[best_code]
            if diff_fact_score < self._conver_prob and diff_score < self._conver_prob:
                break
            last_score = best_fact_score.values[0]

        if is_best:
            best_code = pd.Series(fact_se).sort_values(
                ascending=False)[0:1].index[0]
            best_list = liv_group[best_code]
            best_field = cols[[True if i > 0 else False for i in best_list]]
            return best_code, best_field
        else:
            field_group = {}
            for k, g in liv_group.items():
                field_group[k] = cols[[True if i > 0 else False for i in g]]
            return field_group


# -*- encoding:utf-8 -*-
"""
    日内滑点卖出价格决策基础模块：暂时迁移简单实现方式，符合回测需求，如迁移实盘模块
    需添加日内择时策略，通过日内分钟k线，实现日内分钟k线择时，更微观的
    实现日内择时滑点功能，不考虑大资金的冲击成本及系统外的大幅滑点
"""

from abc import ABCMeta, abstractmethod
import functools

import numpy as np
from ultron.ump.core.fixes import six


class SlippageSellBase(six.with_metaclass(ABCMeta, object)):
    """非高频日内滑点卖出决策抽象基类"""

    def __init__(self, kl_pd_sell, factor_name):
        """
        :param kl_pd_sell: 交易当日的交易数据
        :param factor_name: ABuFactorSellBase实例对象的factor_name
        """
        self.sell_price = np.inf
        self.kl_pd_sell = kl_pd_sell
        self.factor_name = factor_name

    def fit(self):
        """做基础验证比如今天是否停盘后调用fit_price"""
        if self.kl_pd_sell.empty or self.kl_pd_sell.volume == 0:
            # 卖出时负无穷为放弃单子
            return -np.inf

        return self.fit_price()

    @abstractmethod
    def fit_price(self):
        """
        子类主要需要实现的函数，决策交易当日的最终卖出价格
        :return: 最终决策的当前交易卖出价格
        """
        pass


"""是否开启跌停板滑点卖出价格特殊处理，默认关闭，外部修改如：slippage.ssb.g_enable_limit_down = True"""
g_enable_limit_down = False
"""
    初始设定跌停板卖出成交概率100%，这里也可以在计算完一次概率后，再使用成交量做二次概率计算，
    外部修改如：slippage.ssb.g_limit_down_deal_chance = 0.5，即修改为50%成功卖出概率
"""
g_limit_down_deal_chance = 1
"""在集合竞价阶段价格已达成跌停的情况下卖出成功的概率，默认0.2, 即20%成功概率"""
g_pre_limit_down_rate = 0.2


def slippage_limit_down(func):
    """
        针对a股跌停板卖出价格决策的装饰器，子类可选择装饰与不装饰在fit_price上
        如果是实盘策略中，使用分钟k线，及日内择时策略子策略，即不需特别处理。
        回测中需要特别处理，处理卖出成功概率，根据概率决定是否能卖出，
        及跌停下的卖出价格决策，跌停下卖出价格模型为，越靠近跌停板价格
        卖出成交概率越大, 即在跌停下预期以靠近跌停价格卖出，缺点是使用了随机数，
        导致回测结果将出现不一致的情况
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if g_enable_limit_down and self.kl_pd_sell.p_change <= -10 and self.kl_pd_sell.low == self.kl_pd_sell.close:
            """
                跌停板命中后需要根据跌停板买入成交概率(g_limit_down_deal_chance)来作为
                二项式分布的概率值计算卖出成功概率
            """

            if self.kl_pd_sell.high == self.kl_pd_sell.low:
                # 10个点，且最高＝最低，即a股在集合竞价阶段达成跌停，卖出成功概率降低到g_limit_down_deal_chance * 0.2
                # TODO 这个概率最好使用成交量当日来计算出来
                limit_down_deal_chance = g_limit_down_deal_chance * g_pre_limit_down_rate
            else:
                limit_down_deal_chance = g_limit_down_deal_chance

            deal = np.random.binomial(1, limit_down_deal_chance)
            if deal:
                # 卖出成功后需要进一步决策价位，首选arange出一个从高到跌停价格的序列，间隔-0.01
                if self.kl_pd_sell.high == self.kl_pd_sell.low:
                    return self.kl_pd_sell.low

                price_hl = np.arange(self.kl_pd_sell.high, self.kl_pd_sell.low,
                                     -0.01)
                hl_chance = np.linspace(0, 1, len(price_hl))
                # noinspection PyUnresolvedReferences
                p = hl_chance / hl_chance.sum()
                # 最后使用随机加权概率抽取，选中一个卖出价格
                return np.random.choice(price_hl, 1, p=p)[0]
            # 没能成交返回负无穷
            return -np.inf
        else:
            return func(self, *args, **kwargs)

    return wrapper



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd


def qs_percentile(data, universe, lower, upper):
    # rval: of type pd.DataFrame
    rval = data.copy()
    nrow, ncol = rval.shape
    rval[~universe] = np.nan
    # qlow, qup: of type pd.Series
    qlow = rval.quantile(q=lower, axis=1)
    qup = rval.quantile(q=upper, axis=1)
    # qlow, qup: of type pd.DataFrame
    qlow = qlow.repeat(ncol).values.reshape((nrow, ncol))
    qlow = pd.DataFrame(qlow, index=rval.index, columns=rval.columns)
    qup = qup.repeat(ncol).values.reshape((nrow, ncol))
    qup = pd.DataFrame(qup, index=rval.index, columns=rval.columns)

    blo = rval >= qlow
    bup = rval <= qup
    idx = blo & bup
    rval[~idx] = np.nan
    return rval


def cross_section_universe_equal(universe):
    rval = universe.div(np.sum(universe, axis=1) + 1e-16, axis=0)
    return rval


def cross_section_nan_scale(data):
    data = data.copy()
    denominator = data.abs().sum(axis=1, skipna=True)
    rval = data.div(denominator, axis=0)
    return rval


def cross_section_rank(data):
    data = data.copy()
    rval = data.rank(axis=1,
                     method='average',
                     na_option='keep',
                     ascending=True,
                     pct=False)
    return rval


def cross_section_scale(data):
    data = data.copy()
    denominator = data.abs().sum(axis=1, skipna=True)
    rval = data.div(denominator, axis=0)
    rval.fillna(0, inplace=True)
    return rval


def cross_section_mean_deviation(data, universe):
    data = data.copy()
    data[~universe] = np.nan
    rval = data.subtract(data.mean(axis=1, skipna=True), axis=0)
    return rval


def cross_section_demean_rank(data, universe):
    data = data.copy()
    data[~universe] = np.nan
    rank = data.rank(axis=1,
                     method='average',
                     na_option='keep',
                     ascending=True,
                     pct=False)
    rval = cross_section_mean_deviation(rank, universe)
    return rval


def ts_grade(data, step):
    raw_in = data.copy()
    if isinstance(data, pd.Series):
        data = pd.DataFrame(data=data)
    nrow, ncol = data.shape
    x = data.copy()
    flag = x.isnull()
    x[flag] = 0
    x = x.values

    rval = np.full(data.shape, 0.0)
    for i in range(1, nrow):
        for j in range(ncol):
            multiple = x[i, j] / step
            multiple_round = np.round(multiple + 1e-16 * np.sign(multiple))
            multiple_ceil = np.ceil(multiple)
            if multiple_round != multiple and x[i, j] > rval[i - 1, j]:
                rval[i, j] = (multiple_ceil - 1) * step
            else:
                rval[i, j] = multiple_ceil * step
    if isinstance(raw_in, pd.DataFrame):
        rval = pd.DataFrame(data=rval,
                            index=raw_in.index,
                            columns=raw_in.columns)
    elif isinstance(raw_in, pd.Series):
        rval = pd.Series(data=rval.reshape(len(rval), ), index=raw_in.index)
    rval[flag] = np.nan
    return rval


def ts_delay(data, winsize):
    rval = data.shift(periods=winsize, freq=None, axis=0)
    return rval


def ts_true_high(c, hi):
    # c: close
    # hi: high
    c_delayed = ts_delay(c, 1)
    factor = np.maximum(c_delayed.values, hi.values)
    rval = pd.DataFrame(factor)
    rval.index = c.index
    rval.columns = c.columns
    return rval


def ts_true_low(c, lo):
    # c: close
    # lo: low
    c_delayed = ts_delay(c, 1)
    factor = np.minimum(c_delayed.values, lo.values)
    rval = pd.DataFrame(factor)
    rval.index = c.index
    rval.columns = c.columns
    return rval


def ts_true_range(c, hi, lo):

    true_high = ts_true_high(c, hi)
    true_low = ts_true_low(c, lo)
    rval = true_high - true_low
    return rval


def ts_std_mean0(data, winsize):
    rval = data.pow(2).rolling(window=winsize, center=False, axis=0).mean()
    rval = rval.pow(1 / 2)
    return rval


def ts_zscore_normalization(data, bound=None, winsize=None):

    def get_plus(ser):
        if sum(ser.isnull()) == 0:
            _plus = 0
        else:
            _plus = ser.index.get_loc(ser[ser.isnull()].index[-1]) + 1
        return _plus

    nrow, ncol = data.shape
    rval = np.full((nrow, ncol), np.nan)

    plus_series = data.apply(get_plus).tolist()
    for j, col in enumerate(data):
        plus = plus_series[j]
        if winsize + plus > nrow:
            continue

        x = data[col].values

        x_truncated = np.zeros_like(x)
        x_truncated[plus + 1:winsize + plus + 1] = x[plus + 1:winsize + plus +
                                                     1]

        std_ = np.zeros_like(x)

        for i in range(winsize + plus, nrow):
            if i == winsize + plus:
                val = x_truncated[i - winsize + 1:i + 1]
                std0 = np.sqrt(np.mean(val**2))
                val_x_outlier = val[np.abs(val) <=
                                    bound * std0]  # FIXME why not clip
                # val_x_outlier = np.clip(val, -bound * std0, bound * std0)
                std_x_outlier = np.sqrt(np.mean(val_x_outlier**2))

                threshold = bound * std_x_outlier
                np.clip(val, -threshold, threshold, val)

                std_[i] = np.sqrt(np.mean(val**2))
                x_truncated[i - winsize + 1:i + 1] = val
            else:
                std_m1 = std_[i - 1]
                if std_m1 == 0:
                    std_m1 = 1e-7
                x_truncated[i] = np.clip(x[i], -bound * std_m1, bound * std_m1)
                var_ = std_m1**2 * winsize - x_truncated[
                    i - winsize]**2 + x_truncated[i]**2

                if var_ < 0:
                    var_ = 0
                std_[i] = np.sqrt(var_ / winsize)

                rval[i, j] = x_truncated[i] / std_m1
    rval = rval / bound
    rval = pd.DataFrame(data=rval, index=data.index, columns=data.columns)
    return rval



# -*- coding: utf-8 -*-
import numpy as np
import time, datetime, hashlib, copy
from ultron.utilities.logger import kd_logger
from .operators import crossover_sets, mutation_sets, calc_factor, Function, FunctionType

import warnings

warnings.filterwarnings("ignore")

ABS_FLOAT = 0.000001


class Program(object):

    def __init__(self,
                 init_depth,
                 method,
                 random_state,
                 factor_sets,
                 p_point_replace,
                 function_set,
                 operators_set,
                 gen,
                 fitness,
                 coverage_rate=0.8,
                 n_features=0,
                 program=None,
                 parents=None):
        self._init_depth = init_depth
        self._init_method = method
        self._program = program
        self._factor_sets = factor_sets
        self._p_point_replace = p_point_replace
        self._function_set = function_set
        self._operators_set = operators_set
        self._n_features = n_features
        self._fitness = fitness
        self._coverage_rate = coverage_rate
        self._gen = gen
        self._raw_fitness = None  # fitness得分
        self._is_valid = True
        self._parents = parents
        self._retain_data = None
        self._create_time = datetime.datetime.now()
        self._name = 'ultron_' + str(
            int(time.time() * 1000000 + datetime.datetime.now().microsecond))
        if self._program is None:
            self._program = self.build_program(random_state)
        self.variation()
        self.create_identification()

    def parent_idx(self):
        return 0 if self._parents is None else self._parents['parent_idx']

    def log(self):
        parents = {'method': 'gen'} if self._parents is None else self._parents
        formual = self.transform()
        identification = self._identification
        name = self._name
        kd_logger.info(
            'name:%s,method:%s,gen:%d,formual:%s,fitness:%f,identification:%s'
            % (name, str(parents['method']), self._gen, formual,
               self._raw_fitness, identification))

    def expression(self):
        res = []
        for node in self._program:
            if isinstance(node, Function):
                res.append(node.name)
            else:
                res.append(node)
        return res

    def output(self):
        parents = {'method': 'Gen'} if self._parents is None else self._parents
        return {
            'name': self._name,
            'method': parents['method'],
            'gen': self._gen,
            'features': self._identification,
            'formual': self.transform(),
            'fitness': self._raw_fitness,
            'update_time': self._create_time
        }

    # 交叉变异时会出生成无效子代,设置无效标识
    # 如 ['CurrentAssetsTRate', 'CurrentAssetsTRate', 'rskew_std']
    def variation(self):
        check_result = [p for p in self._program if isinstance(p, Function)]
        if len(check_result) == 0 and len(self._program) > 1:
            self._program = self._program[:1]
            self._is_valid = False

    def create_identification(self):
        m = hashlib.md5()
        try:
            token = self.transform()
        except Exception as e:
            #ID为key
            token = self._name
        if token is None:
            token = self._name
        m.update(bytes(token, encoding='UTF-8'))
        self._identification = m.hexdigest()

    def create_formual(self, apply_formual):
        function = apply_formual[0]
        formula = function.function.__name__
        if function.ftype == FunctionType.cross_section:
            formula += '('
        else:
            formula += ('(' + str(function.default_value) + ',')
        for i in range(0, function.arity):
            if i != 0:
                formula += ','
            if apply_formual[i + 1] in self._factor_sets:
                formula += '\'' + apply_formual[i + 1] + '\''
            else:
                formula += apply_formual[i + 1]
        formula += ')'
        return formula

    def transform(self):
        if len(self._program) < 2:
            result = 'CURRENT(\'' + self._program[0] + '\')'
            return result
        apply_stack = []
        for node in self._program:
            if isinstance(node, Function):
                apply_stack.append([node])
            else:
                try:
                    apply_stack[-1].append(node)
                except Exception as e:
                    return None
            while len(apply_stack[-1]) == apply_stack[-1][0].arity + 1:
                result = self.create_formual(apply_stack[-1])
                if len(apply_stack) != 1:
                    apply_stack.pop()
                    apply_stack[-1].append(result)
                else:
                    return result

    def export_graphviz(self):
        fade_nodes = None
        terminals = []
        if fade_nodes is None:
            fade_nodes = []
        output = 'digraph program {\nnode [style=filled]\n'
        for i, node in enumerate(self._program):
            fill = '#cecece'
            if isinstance(node, Function):
                if i not in fade_nodes:
                    fill = '#2a5caa'
                terminals.append([node.arity, i])
                output += ('%d [label="%s", fillcolor="%s"] ;\n' %
                           (i, node.function.__name__, fill))
            else:
                if i not in fade_nodes:
                    fill = '#60a6f6'
                if node in self._factor_sets:
                    feature_name = node
                else:
                    feature_name = 'X%s' % node
                output += ('%d [label="%s", fillcolor="%s"] ;\n' %
                           (i, feature_name, fill))

                if i == 0:
                    output += '}'
                    return output
                terminals[-1][0] -= 1
                terminals[-1].append(i)
                while terminals[-1][0] == 0:
                    output += '%d -> %d ;\n' % (terminals[-1][1],
                                                terminals[-1][-1])
                    terminals[-1].pop()
                    if len(terminals[-1]) == 2:
                        parent = terminals[-1][-1]
                        terminals.pop()
                        if not terminals:
                            output += '}'
                            return output
                        terminals[-1].append(parent)
                        terminals[-1][0] -= 1

    def build_program(self, random_state):
        #在范围内选取树形深度
        if self._init_method == 'half and half':
            method = ('full' if random_state.randint(2) else 'grow')
        else:
            method = self._init_method
        if isinstance(self._init_depth, int):
            max_depth = self._init_depth
        else:
            max_depth = random_state.randint(*self._init_depth)
        function = random_state.randint(len(self._operators_set))
        function = self._operators_set[function]
        program = [function]
        terminal_stack = [function.arity]
        while terminal_stack:
            depth = len(terminal_stack)
            choice = self._n_features + len(self._operators_set)
            choice = np.random.randint(0, choice)
            if depth < max_depth and (method == 'full'
                                      or choice <= len(self._operators_set)):
                function = self._operators_set[np.random.randint(
                    0,
                    len(self._operators_set) - 1)]
                program.append(function)
                terminal_stack.append(function.arity)
            else:
                factor = self._factor_sets[np.random.randint(
                    0,
                    len(self._factor_sets) - 1)]
                program.append(factor)
                terminal_stack[-1] -= 1
                while terminal_stack[-1] == 0:
                    terminal_stack.pop()
                    if not terminal_stack:
                        return program
                    terminal_stack[-1] -= 1
        return program

    def get_subtree(self, random_state, program=None):
        if program is None:
            program = self._program
        # Choice of crossover points follows Koza's (1992) widely used approach
        # of choosing functions 90% of the time and leaves 10% of the time.
        probs = np.array(
            [0.9 if node in self._operators_set else 0.1 for node in program])
        probs = np.cumsum(probs / probs.sum())
        start = np.searchsorted(probs, random_state.uniform())
        stack = 1
        end = start
        while stack > end - start:
            node = program[end]
            if node in self._operators_set:
                stack += node.arity
            end += 1
        return start, end

    ##复制
    def reproduce(self):
        return copy.deepcopy(self._program)

    ##交叉
    def crossover(self, donor, random_state):
        start, end = self.get_subtree(random_state)
        end -= 1
        removed = range(start, end)
        donor_start, donor_end = self.get_subtree(random_state, donor)
        donor_removed = list(
            set(range(len(donor))) - set(range(donor_start, donor_end)))
        return (self._program[:start] + donor[donor_start:donor_end] +
                self._program[end:]), removed, donor_removed

    ##树变异
    def subtree_mutation(self, random_state):
        chicken = self.build_program(random_state)
        return self.crossover(chicken, random_state)

    ##突变异
    def hoist_mutation(self, random_state):
        start, end = self.get_subtree(random_state)
        subtree = self._program[start:end]
        sub_start, sub_end = self.get_subtree(random_state, subtree)
        hoist = subtree[sub_start:sub_end]
        removed = list(
            set(range(start, end)) -
            set(range(start + sub_start, start + sub_end)))
        return self._program[:start] + hoist + self._program[end:], removed

    ##点变异
    def point_mutation(self, random_state):
        program = copy.deepcopy(self._program)
        mutate = np.where(
            random_state.uniform(size=len(program)) < self._p_point_replace)[0]

        for node in mutate:
            if program[node] in self._operators_set:
                activy = program[node].arity
                #找到参数个数替换
                if activy == 1:
                    replace_node = mutation_sets[np.random.randint(
                        0,
                        len(mutation_sets) - 1)]
                else:
                    replace_node = crossover_sets[np.random.randint(
                        0,
                        len(crossover_sets) - 1)]
                program[node] = replace_node
            else:
                factor = self._factor_sets[np.random.randint(
                    0,
                    len(self._factor_sets) - 1)]
                program[node] = factor
        return program, list(mutate)

    def raw_fitness(self,
                    total_data,
                    factor_sets,
                    default_value,
                    backup_cycle,
                    custom_params,
                    indexs=['trade_date'],
                    key='code'):
        #计算因子值
        if not self._is_valid:
            self._raw_fitness = default_value
            return
        try:
            expression = self.transform()
            if expression is None:
                self._raw_fitness = default_value
                self._is_valid = False
            else:
                factor_data = calc_factor(expression, total_data, indexs, key)
                #切割掉备份周期
                factor_data = factor_data.replace([np.inf, -np.inf], np.nan)
                #处理因子暴露度为0
                factor_data['transformed'] = np.where(
                    np.abs(factor_data.transformed.values) > 0.000001,
                    factor_data.transformed.values, np.nan)
                factor_data = factor_data.loc[factor_data.index.unique()
                                              [backup_cycle:]]
                ##检测覆盖率
                coverage_rate = 1 - factor_data['transformed'].isna().sum(
                ) / len(factor_data['transformed'])
                self._raw_fitness = default_value
                if coverage_rate > self._coverage_rate:
                    cycle_total_data = total_data.copy().set_index(
                        'trade_date')
                    cycle_total_data = cycle_total_data.loc[
                        cycle_total_data.index.unique()[backup_cycle:]]
                    new_custom_params = copy.deepcopy(custom_params)
                    new_custom_params['name'] = self._name
                    results = self._fitness(factor_data,
                                            cycle_total_data.reset_index(),
                                            factor_sets, new_custom_params,
                                            default_value)
                    if isinstance(results, tuple):
                        raw_fitness, self._retain_data = results
                    else:
                        raw_fitness = results

                    self._raw_fitness = default_value if np.isnan(
                        raw_fitness) else raw_fitness

                if self._raw_fitness == default_value:
                    self._is_valid = False

        except Exception as e:
            self._raw_fitness = default_value
            self._is_valid = False
            kd_logger.exception(e)


# -*- coding: utf-8 -*-

from . othgnz_method import schmidt, canonial, symmetry,orthogonalize 

func_dict = {'schmidt':schmidt, 'canonial':canonial, 'symmetry':symmetry, 'onalize':orthogonalize}
   
class OthgnzEngine(object):
    @classmethod
    def create_engine(cls, ce_name):
        return func_dict[ce_name]


from ultron.factor.dimension.corrcoef import Corrcoef as DimensionCorrcoef
from ultron.factor.dimension.lda import LDA as DimensionLDA
from ultron.factor.dimension.pca import PCA as DimensionPCA
from ultron.factor.dimension.kmeans import KMeans as DimensionKeans

__all__ = [
    'DimensionCorrcoef', 'DimensionLDA', 'DimensionPCA', 'DimensionKeans'
]



# -*- encoding:utf-8 -*-
"""

"""

from ultron.ump.indicator import atr
from ultron.ump.indicator import boll
from ultron.ump.indicator import ma
from ultron.ump.indicator import macd
from ultron.ump.indicator import rsi

__all__ = ['atr', 'boll', 'ma', 'macd', 'rsi']



# -*- encoding:utf-8 -*-
"""
    卖出择时示例因子，双均线策略
"""

from ultron.ump.factor.sell.base import FactorSellXD, ESupportDirection
from ultron.ump.indicator.ma import calc_ma_from_prices


class DoubleMaSell(FactorSellXD):
    """示例卖出双均线择时因子"""

    def _init_self(self, **kwargs):
        """
            kwargs中可选参数：fast: 均线快线周期，默认不设置，使用5
            kwargs中可选参数：slow: 均线慢线周期，默认不设置，使用60
        """
        # TODO 重构与买入因子重复代码抽取
        # 均线快线周期，默认使用5天均线
        self.ma_fast = kwargs.pop('fast', 5)
        # 均线慢线周期，默认使用60天均线
        self.ma_slow = kwargs.pop('slow', 60)

        if self.ma_fast >= self.ma_slow:
            # 慢线周期必须大于快线
            raise ValueError('ma_fast >= self.ma_slow !')

        # xd周期数据需要比ma_slow大一天，这样计算ma就可以拿到今天和昨天两天的ma，用来判断金叉，死叉
        kwargs['xd'] = self.ma_slow + 1
        # 设置好xd后可以直接使用基类针对xd的初始化
        super(DoubleMaSell, self)._init_self(**kwargs)

    def support_direction(self):
        """支持的方向，因子支持两个方向"""
        return [
            ESupportDirection.DIRECTION_CAll.value,
            ESupportDirection.DIRECTION_PUT.value
        ]

    def fit_day(self, today, orders):
        """
            双均线卖出择时因子：
            call方向：快线下穿慢线形成死叉，做为卖出信号
            put方向： 快线上穿慢线做为卖出信号
        """
        # 计算快线
        fast_line = calc_ma_from_prices(self.xd_kl.close,
                                        self.ma_fast,
                                        min_periods=1)
        # 计算慢线
        slow_line = calc_ma_from_prices(self.xd_kl.close,
                                        self.ma_slow,
                                        min_periods=1)

        if len(fast_line) >= 2 and len(slow_line) >= 2:
            # 今天的快线值
            fast_today = fast_line[-1]
            # 昨天的快线值
            fast_yesterday = fast_line[-2]
            # 今天的慢线值
            slow_today = slow_line[-1]
            # 昨天的慢线值
            slow_yesterday = slow_line[-2]

            for order in orders:
                if order.expect_direction == 1 \
                        and fast_yesterday >= slow_yesterday and fast_today < slow_today:
                    # call方向：快线下穿慢线线形成死叉，做为卖出信号
                    self.sell_tomorrow(order)
                elif order.expect_direction == -1 \
                        and slow_yesterday >= fast_yesterday and fast_today > slow_today:
                    # put方向：快线上穿慢线做为卖出信号
                    self.sell_tomorrow(order)